import enum
import inspect

from dataclasses import dataclass
from collections import defaultdict
from typing import List, Dict, Optional

from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseChatModel
from mindsdb_sql_parser.ast import Select, BinaryOperation, Identifier, Constant, Star

from mindsdb.utilities import log
from mindsdb.utilities.cache import get_cache
from mindsdb.utilities.config import config
from mindsdb.interfaces.storage import db
from mindsdb.interfaces.skills.sql_agent import SQLAgent
from mindsdb.integrations.libs.vectordatabase_handler import TableField
from mindsdb.interfaces.agents.constants import DEFAULT_TEXT2SQL_DATABASE

_DEFAULT_TOP_K_SIMILARITY_SEARCH = 5
_MAX_CACHE_SIZE = 1000

logger = log.getLogger(__name__)


class SkillType(enum.Enum):
    TEXT2SQL_LEGACY = "text2sql"
    TEXT2SQL = "sql"
    KNOWLEDGE_BASE = "knowledge_base"
    RETRIEVAL = "retrieval"


@dataclass
class SkillData:
    """Storage for skill's data

    Attributes:
        name (str): name of the skill
        type (str): skill's type (SkillType)
        params (dict): skill's attributes
        project_id (int): id of the project
        agent_tables_list (Optional[List[str]]): the restriction on available tables for an agent using the skill
    """

    name: str
    type: str
    params: dict
    project_id: int
    agent_tables_list: Optional[List[str]]

    @property
    def restriction_on_tables(self) -> Optional[Dict[str, set]]:
        """Schemas and tables which agent+skill may use. The result is intersections of skill's and agent's tables lists.

        Returns:
            Optional[Dict[str, set]]: allowed schemas and tables. Schemas - are keys in dict, tables - are values.
                if result is None, then there are no restrictions

        Raises:
            ValueError: if there is no intersection between skill's and agent's list.
                This means that all tables restricted for use.
        """

        def list_to_map(input: List) -> Dict:
            agent_tables_map = defaultdict(set)
            for x in input:
                if isinstance(x, str):
                    table_name = x
                    schema_name = None
                elif isinstance(x, dict):
                    table_name = x["table"]
                    schema_name = x.get("schema")
                else:
                    raise ValueError(f"Unexpected value in tables list: {x}")
                agent_tables_map[schema_name].add(table_name)
            return agent_tables_map

        agent_tables_map = list_to_map(self.agent_tables_list or [])
        skill_tables_map = list_to_map(self.params.get("tables", []))

        if len(agent_tables_map) > 0 and len(skill_tables_map) > 0:
            if len(set(agent_tables_map) & set(skill_tables_map)) == 0:
                raise ValueError("Skill's and agent's allowed tables list have no shared schemas.")

            intersection_tables_map = defaultdict(set)
            has_intersection = False
            for schema_name in agent_tables_map:
                if schema_name not in skill_tables_map:
                    continue
                intersection_tables_map[schema_name] = agent_tables_map[schema_name] & skill_tables_map[schema_name]
                if len(intersection_tables_map[schema_name]) > 0:
                    has_intersection = True
            if has_intersection is False:
                raise ValueError("Skill's and agent's allowed tables list have no shared tables.")
            return intersection_tables_map
        if len(skill_tables_map) > 0:
            return skill_tables_map
        if len(agent_tables_map) > 0:
            return agent_tables_map
        return None


class SkillToolController:
    def __init__(self):
        self.command_executor = None

    def get_command_executor(self):
        if self.command_executor is None:
            from mindsdb.api.executor.command_executor import ExecuteCommands
            from mindsdb.api.executor.controllers import (
                SessionController,
            )  # Top-level import produces circular import in some cases TODO: figure out a fix without losing runtime improvements (context: see #9304)  # noqa

            sql_session = SessionController()
            sql_session.database = config.get("default_project")

            self.command_executor = ExecuteCommands(sql_session)
        return self.command_executor

    def _make_text_to_sql_tools(self, skills: List[db.Skills], llm) -> List:
        """
        Uses SQLAgent to execute tool
        """
        # To prevent dependency on Langchain unless an actual tool uses it.
        try:
            from mindsdb.interfaces.agents.mindsdb_database_agent import MindsDBSQL
            from mindsdb.interfaces.skills.custom.text2sql.mindsdb_sql_toolkit import MindsDBSQLToolkit
        except ImportError:
            raise ImportError(
                "To use the text-to-SQL skill, please install langchain with `pip install mindsdb[langchain]`"
            )

        command_executor = self.get_command_executor()

        def escape_table_name(name: str) -> str:
            name = name.strip(" `")
            return f"`{name}`"

        tables_list = []
        knowledge_bases_list = []
        ignore_knowledge_bases_list = []

        # Track databases extracted from dot notation
        extracted_databases = set()

        # Initialize knowledge_base_database with default value
        knowledge_base_database = DEFAULT_TEXT2SQL_DATABASE  # Default to mindsdb project

        # First pass: collect all database and knowledge base parameters
        for skill in skills:
            # Update knowledge_base_database if specified in any skill
            if skill.params.get("knowledge_base_database"):
                knowledge_base_database = skill.params.get("knowledge_base_database")

            # Extract databases from include_tables with dot notation
            if skill.params.get("include_tables"):
                include_tables = skill.params.get("include_tables")
                if isinstance(include_tables, str):
                    include_tables = [t.strip() for t in include_tables.split(",")]

                # Extract database names from dot notation
                for table in include_tables:
                    if "." in table:
                        db_name = table.split(".")[0]
                        extracted_databases.add(db_name)

            # Extract databases from include_knowledge_bases with dot notation
            if skill.params.get("include_knowledge_bases"):
                include_kbs = skill.params.get("include_knowledge_bases")
                if isinstance(include_kbs, str):
                    include_kbs = [kb.strip() for kb in include_kbs.split(",")]

                # Extract database names from dot notation
                for kb in include_kbs:
                    if "." in kb:
                        db_name = kb.split(".")[0]
                        if db_name != knowledge_base_database:
                            # Only update if it's different from the default
                            knowledge_base_database = db_name

        # Second pass: collect all tables and knowledge base restrictions
        for skill in skills:
            # Get database for tables (this is an actual database connection)
            database = skill.params.get("database", DEFAULT_TEXT2SQL_DATABASE)

            # Add databases extracted from dot notation if no explicit database is provided
            if not database and extracted_databases:
                # Use the first extracted database if no explicit database is provided
                database = next(iter(extracted_databases))
                # Update the skill params with the extracted database
                skill.params["database"] = database

            # Extract knowledge base restrictions if they exist in the skill params
            if skill.params.get("include_knowledge_bases"):
                # Convert to list if it's a string
                include_kbs = skill.params.get("include_knowledge_bases")
                if isinstance(include_kbs, str):
                    include_kbs = [kb.strip() for kb in include_kbs.split(",")]

                # Process each knowledge base name
                for kb in include_kbs:
                    # If it doesn't have a dot, prefix it with the knowledge_base_database
                    if "." not in kb:
                        knowledge_bases_list.append(f"{knowledge_base_database}.{kb}")
                    else:
                        knowledge_bases_list.append(kb)

            # Collect ignore_knowledge_bases
            if skill.params.get("ignore_knowledge_bases"):
                # Convert to list if it's a string
                ignore_kbs = skill.params.get("ignore_knowledge_bases")
                if isinstance(ignore_kbs, str):
                    ignore_kbs = [kb.strip() for kb in ignore_kbs.split(",")]

                # Process each knowledge base name to ignore
                for kb in ignore_kbs:
                    # If it doesn't have a dot, prefix it with the knowledge_base_database
                    if "." not in kb:
                        ignore_knowledge_bases_list.append(f"{knowledge_base_database}.{kb}")
                    else:
                        ignore_knowledge_bases_list.append(kb)

            # Skip if no database specified
            if not database:
                continue

            # Process include_tables with dot notation
            if skill.params.get("include_tables"):
                include_tables = skill.params.get("include_tables")
                if isinstance(include_tables, str):
                    include_tables = [t.strip() for t in include_tables.split(",")]

                for table in include_tables:
                    # If table already has a database prefix, use it as is
                    if "." in table:
                        # Check if the table already has backticks
                        if "`" in table:
                            tables_list.append(table)
                        else:
                            # Apply escape_table_name only to the table part
                            parts = table.split(".")
                            if len(parts) == 2:
                                # Format: database.table
                                tables_list.append(f"{parts[0]}.{escape_table_name(parts[1])}")
                            elif len(parts) == 3:
                                # Format: database.schema.table
                                tables_list.append(f"{parts[0]}.{parts[1]}.{escape_table_name(parts[2])}")
                            else:
                                # Unusual format, escape the whole thing
                                tables_list.append(escape_table_name(table))
                    else:
                        # Otherwise, prefix with the database
                        tables_list.append(f"{database}.{escape_table_name(table)}")

                # Skip further table processing if include_tables is specified
                continue

            restriction_on_tables = skill.restriction_on_tables

            if restriction_on_tables is None and database:
                try:
                    handler = command_executor.session.integration_controller.get_data_handler(database)
                    if "all" in inspect.signature(handler.get_tables).parameters:
                        response = handler.get_tables(all=True)
                    else:
                        response = handler.get_tables()
                    # no restrictions
                    columns = [c.lower() for c in response.data_frame.columns]
                    name_idx = columns.index("table_name") if "table_name" in columns else 0

                    if "table_schema" in response.data_frame.columns:
                        for _, row in response.data_frame.iterrows():
                            tables_list.append(f"{database}.{row['table_schema']}.{escape_table_name(row[name_idx])}")
                    else:
                        for table_name in response.data_frame.iloc[:, name_idx]:
                            tables_list.append(f"{database}.{escape_table_name(table_name)}")
                except Exception as e:
                    logger.warning(f"Could not get tables from database {database}: {str(e)}")
                continue

            # Handle table restrictions
            if restriction_on_tables and database:
                for schema_name, tables in restriction_on_tables.items():
                    for table in tables:
                        # Check if the table already has dot notation (e.g., 'postgresql_conn.home_rentals')
                        if "." in table:
                            # Table already has database prefix, add it directly
                            tables_list.append(escape_table_name(table))
                        else:
                            # No dot notation, apply schema and database as needed
                            if schema_name is None:
                                tables_list.append(f"{database}.{escape_table_name(table)}")
                            else:
                                tables_list.append(f"{database}.{schema_name}.{escape_table_name(table)}")
                continue

        # Remove duplicates from lists
        tables_list = list(set(tables_list))
        knowledge_bases_list = list(set(knowledge_bases_list))
        ignore_knowledge_bases_list = list(set(ignore_knowledge_bases_list))

        # Determine knowledge base parameters to pass to SQLAgent
        include_knowledge_bases = knowledge_bases_list if knowledge_bases_list else None
        ignore_knowledge_bases = ignore_knowledge_bases_list if ignore_knowledge_bases_list else None

        # If both include and ignore lists exist, include takes precedence
        if include_knowledge_bases:
            ignore_knowledge_bases = None

        # # Get all databases from skills and extracted databases
        # all_databases = list(set([s.params.get('database', DEFAULT_TEXT2SQL_DATABASE) for s in skills if s.params.get('database')] + list(extracted_databases)))
        #
        #
        # # If no databases were specified or extracted, use 'mindsdb' as a default
        # if not all_databases:
        #     all_databases = [DEFAULT_TEXT2SQL_DATABASE]
        #

        all_databases = []
        # Filter out None values
        all_databases = [db for db in all_databases if db is not None]

        # Create a databases_struct dictionary that includes all extracted databases
        databases_struct = {}

        # First, add databases from skills with explicit database parameters
        for skill in skills:
            if skill.params.get("database"):
                databases_struct[skill.params["database"]] = skill.restriction_on_tables

        # Then, add all extracted databases with no restrictions
        for db_name in extracted_databases:
            if db_name not in databases_struct:
                databases_struct[db_name] = None

        sql_agent = SQLAgent(
            command_executor=command_executor,
            databases=all_databases,
            databases_struct=databases_struct,
            include_tables=tables_list,
            ignore_tables=None,
            include_knowledge_bases=include_knowledge_bases,
            ignore_knowledge_bases=ignore_knowledge_bases,
            knowledge_base_database=knowledge_base_database,
            sample_rows_in_table_info=3,
            cache=get_cache("agent", max_size=_MAX_CACHE_SIZE),
        )
        db = MindsDBSQL.custom_init(sql_agent=sql_agent)
        should_include_kb_tools = include_knowledge_bases is not None and len(include_knowledge_bases) > 0
        should_include_tables_tools = len(databases_struct) > 0 or len(tables_list) > 0
        toolkit = MindsDBSQLToolkit(
            db=db,
            llm=llm,
            include_tables_tools=should_include_tables_tools,
            include_knowledge_base_tools=should_include_kb_tools,
        )
        return toolkit.get_tools()

    def _make_retrieval_tools(self, skill: db.Skills, llm, embedding_model):
        """
        creates advanced retrieval tool i.e. RAG
        """
        params = skill.params
        config = params.get("config", {})
        if "llm" not in config:
            # Set LLM if not explicitly provided in configs.
            config["llm"] = llm
        tool = dict(
            name=params.get("name", skill.name),
            source=params.get("source", None),
            config=config,
            description=f"You must use this tool to get more context or information "
            f"to answer a question about {params['description']}. "
            f"The input should be the exact question the user is asking.",
            type=skill.type,
        )
        pred_args = {}
        pred_args["llm"] = llm

        from .retrieval_tool import build_retrieval_tools

        return build_retrieval_tools(tool, pred_args, skill)

    def _get_rag_query_function(self, skill: db.Skills):
        session_controller = self.get_command_executor().session

        def _answer_question(question: str) -> str:
            knowledge_base_name = skill.params["source"]

            # make select in KB table
            query = Select(
                targets=[Star()],
                where=BinaryOperation(op="=", args=[Identifier(TableField.CONTENT.value), Constant(question)]),
                limit=Constant(_DEFAULT_TOP_K_SIMILARITY_SEARCH),
            )
            kb_table = session_controller.kb_controller.get_table(knowledge_base_name, skill.project_id)

            res = kb_table.select_query(query)
            # Handle both chunk_content and content column names
            if hasattr(res, "chunk_content"):
                return "\n".join(res.chunk_content)
            elif hasattr(res, "content"):
                return "\n".join(res.content)
            else:
                return "No content or chunk_content found in knowledge base response"

        return _answer_question

    def _make_knowledge_base_tools(self, skill: db.Skills) -> dict:
        # To prevent dependency on Langchain unless an actual tool uses it.
        description = skill.params.get("description", "")

        logger.warning(
            "This skill is deprecated and will be removed in the future. Please use `retrieval` skill instead "
        )

        return dict(
            name="Knowledge Base Retrieval",
            func=self._get_rag_query_function(skill),
            description=f"Use this tool to get more context or information to answer a question about {description}. The input should be the exact question the user is asking.",
            type=skill.type,
        )

    def get_tools_from_skills(
        self, skills_data: List[SkillData], llm: BaseChatModel, embedding_model: Embeddings
    ) -> dict:
        """Creates function for skill and metadata (name, description)

        Args:
            skills_data (List[SkillData]): Skills to make a tool from
            llm (BaseChatModel): LLM which will be used by skills
            embedding_model (Embeddings): this model is used by retrieval skill

        Returns:
            dict: with keys: name, description, func
        """

        # group skills by type
        skills_group = defaultdict(list)
        for skill in skills_data:
            try:
                skill_type = SkillType(skill.type)
            except ValueError:
                raise NotImplementedError(
                    f"skill of type {skill.type} is not supported as a tool, supported types are: {list(SkillType._member_names_)}"
                )

            if skill_type == SkillType.TEXT2SQL_LEGACY:
                skill_type = SkillType.TEXT2SQL
            skills_group[skill_type].append(skill)

        tools = {}
        for skill_type, skills in skills_group.items():
            if skill_type == SkillType.TEXT2SQL:
                tools[skill_type] = self._make_text_to_sql_tools(skills, llm)
            elif skill_type == SkillType.KNOWLEDGE_BASE:
                tools[skill_type] = [self._make_knowledge_base_tools(skill) for skill in skills]
            elif skill_type == SkillType.RETRIEVAL:
                tools[skill_type] = []
                for skill in skills:
                    tools[skill_type] += self._make_retrieval_tools(skill, llm, embedding_model)
        return tools


skill_tool = SkillToolController()



import re
from typing import Dict
import pandas as pd
from pyhive import sqlalchemy_trino
from mindsdb_sql_parser import parse_sql, ASTNode
from trino.auth import BasicAuthentication
from trino.dbapi import connect
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)

logger = log.getLogger(__name__)


class TrinoHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Trino statements

    kerberos is not implemented yet
    """

    name = 'trino'

    def __init__(self, name, connection_data, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.connection_data = connection_data
        '''
        service_name = kwargs.get('service_name')
        self.config_file_name = kwargs.get('config_file_name')
        self.trino_config_provider = TrinoConfigProvider(config_file_name=self.config_file_name)
        self.kerberos_config = self.trino_config_provider.get_trino_kerberos_config()
        self.http_scheme = self.kerberos_config['http_scheme']
        self.dialect = self.kerberos_config['dialect']
        config = self.kerberos_config['config']
        hostname_override = self.kerberos_config['hostname_override']
        principal = f"{kwargs.get('user')}@{hostname_override}"
        ca_bundle = self.kerberos_config['ca_bundle']
        self.auth_config = KerberosAuthentication(config=config,
                                                  service_name=service_name,
                                                  principal=principal,
                                                  ca_bundle=ca_bundle,
                                                  hostname_override=hostname_override)
        '''
        self.connection = None
        self.is_connected = False
        self.with_clause = ""

    def connect(self):
        """"
        Handles the connection to a Trino instance.
        """
        if self.is_connected is True:
            return self.connection

        # option configuration
        http_scheme = 'http'
        auth = None
        auth_config = None
        password = None

        if 'auth' in self.connection_data:
            auth = self.connection_data['auth']
        if 'password' in self.connection_data:
            password = self.connection_data['password']
        if 'http_scheme' in self.connection_data:
            http_scheme = self.connection_data['http_scheme']
        if 'with' in self.connection_data:
            self.with_clause = self.connection_data['with']

        if password and auth == 'kerberos':
            raise Exception("Kerberos authorization doesn't support password.")
        elif password:
            auth_config = BasicAuthentication(self.connection_data['user'], password)

        if auth:
            conn = connect(
                host=self.connection_data['host'],
                port=self.connection_data['port'],
                user=self.connection_data['user'],
                catalog=self.connection_data['catalog'],
                schema=self.connection_data['schema'],
                http_scheme=http_scheme,
                auth=auth_config)
        else:
            conn = connect(
                host=self.connection_data['host'],
                port=self.connection_data['port'],
                user=self.connection_data['user'],
                catalog=self.connection_data['catalog'],
                schema=self.connection_data['schema'])

        self.is_connected = True
        self.connection = conn
        return conn

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the Trino instance
        :return: success status and error message if error occurs
        """
        response = StatusResponse(False)

        try:
            connection = self.connect()
            cur = connection.cursor()
            cur.execute("SELECT 1")
            response.success = True
        except Exception as e:
            logger.error(f'Error connecting to Trino {self.connection_data["schema"]}, {e}!')
            response.error_message = str(e)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in Trino
        :return: returns the records from the current recordset
        """
        try:
            connection = self.connect()
            cur = connection.cursor()
            result = cur.execute(query)
            if result and cur.description:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result,
                        columns=[x[0] for x in cur.description]
                    )
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
            connection.commit()
        except Exception as e:
            logger.error(f'Error connecting to Trino {self.connection_data["schema"]}, {e}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )
        return response

    def query(self, query: ASTNode) -> Response:
        # Utilize trino dialect from sqlalchemy
        # implement WITH clause as default for all table
        # in future, this behavior should be changed to support more detail
        # level
        # also, for simple the current implement is using rendered query string
        # another method that directly manipulate ASTNOde is prefered
        renderer = SqlalchemyRender(sqlalchemy_trino.TrinoDialect)
        query_str = renderer.get_string(query, with_failback=True)
        modified_query_str = re.sub(
            r"(?is)(CREATE.+TABLE.+\(.*\))",
            f"\\1 {self.with_clause}",
            query_str
        )
        return self.native_query(modified_query_str)

    def get_tables(self) -> Response:
        """
        List all tables in Trino
        :return: list of all tables
        """
        query = "SHOW TABLES"
        response = self.native_query(query)
        df = response.data_frame
        response.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return response

    def get_columns(self, table_name: str) -> Dict:
        query = f'DESCRIBE "{table_name}"'
        response = self.native_query(query)
        return response



import pandas as pd
from googleapiclient.discovery import build

from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.libs.api_handler import APIHandler, FuncParser
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities.config import Config
from mindsdb.utilities import log
from mindsdb.integrations.utilities.handlers.auth_utilities.google import GoogleUserOAuth2Manager
from mindsdb.integrations.utilities.handlers.auth_utilities.exceptions import AuthException

from .google_calendar_tables import GoogleCalendarEventsTable

DEFAULT_SCOPES = [
    "https://www.googleapis.com/auth/calendar",
    "https://www.googleapis.com/auth/calendar.events",
    "https://www.googleapis.com/auth/calendar.readonly",
]

logger = log.getLogger(__name__)


class GoogleCalendarHandler(APIHandler):
    """
    A class for handling connections and interactions with the Google Calendar API.
    """

    name = "google_calendar"

    def __init__(self, name: str, **kwargs):
        """constructor
        Args:
            name (str): the handler name
            credentials_file (str): The path to the credentials file.
            scopes (list): The list of scopes to use for authentication.
            is_connected (bool): Whether the API client is connected to Google Calendar.
            events (GoogleCalendarEventsTable): The `GoogleCalendarEventsTable` object for interacting with the events table.
        """
        super().__init__(name)
        self.connection_data = kwargs.get("connection_data", {})

        self.service = None
        self.is_connected = False

        self.handler_storage = kwargs["handler_storage"]

        self.credentials_url = self.connection_data.get("credentials_url", None)
        self.credentials_file = self.connection_data.get("credentials_file", None)
        if self.connection_data.get("credentials"):
            self.credentials_file = self.connection_data.pop("credentials")
        if not self.credentials_file and not self.credentials_url:
            # try to get from config
            gcalendar_config = Config().get("handlers", {}).get("youtube", {})
            secret_file = gcalendar_config.get("credentials_file")
            secret_url = gcalendar_config.get("credentials_url")
            if secret_file:
                self.credentials_file = secret_file
            elif secret_url:
                self.credentials_url = secret_url

        self.scopes = self.connection_data.get("scopes", DEFAULT_SCOPES)

        events = GoogleCalendarEventsTable(self)
        self.events = events
        self._register_table("events", events)

    def connect(self):
        """
        Set up any connections required by the handler
        Should return output of check_connection() method after attempting
        connection. Should switch self.is_connected.
        Returns:
            HandlerStatusResponse
        """
        if self.is_connected is True:
            return self.service

        google_oauth2_manager = GoogleUserOAuth2Manager(
            self.handler_storage,
            self.scopes,
            self.credentials_file,
            self.credentials_url,
            self.connection_data.get("code"),
        )
        creds = google_oauth2_manager.get_oauth2_credentials()

        self.service = build("calendar", "v3", credentials=creds)
        return self.service

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
            response.copy_storage = True

        except AuthException as error:
            response.error_message = str(error)
            response.redirect_url = error.auth_url
            return response

        except Exception as e:
            logger.error(f"Error connecting to Google Calendar API: {e}!")
            response.error_message = e

        self.is_connected = response.success
        return response

    def native_query(self, query: str = None) -> Response:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                api's json etc)
        Returns:
            HandlerResponse
        """
        method_name, params = FuncParser().from_string(query)

        df = self.call_application_api(method_name, params)

        return Response(RESPONSE_TYPE.TABLE, data_frame=df)

    def get_events(self, params: dict = None) -> pd.DataFrame:
        """
        Get events from Google Calendar API
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        page_token = None
        events = pd.DataFrame(columns=self.events.get_columns())
        while True:
            events_result = service.events().list(calendarId="primary", pageToken=page_token, **params).execute()
            events = pd.concat(
                [events, pd.DataFrame(events_result.get("items", []), columns=self.events.get_columns())],
                ignore_index=True,
            )
            page_token = events_result.get("nextPageToken")
            if not page_token:
                break
        return events

    def create_event(self, params: dict = None) -> pd.DataFrame:
        """
        Create an event in the calendar.
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        # Check if 'attendees' is a string and split it into a list
        if isinstance(params["attendees"], str):
            params["attendees"] = params["attendees"].split(",")

        event = {
            "summary": params["summary"],
            "location": params["location"],
            "description": params["description"],
            "start": {
                "dateTime": params["start"]["dateTime"],
                "timeZone": params["start"]["timeZone"],
            },
            "end": {
                "dateTime": params["end"]["dateTime"],
                "timeZone": params["end"]["timeZone"],
            },
            "recurrence": ["RRULE:FREQ=DAILY;COUNT=1"],
            "attendees": [
                {"email": attendee["email"]}
                for attendee in (
                    params["attendees"] if isinstance(params["attendees"], list) else [params["attendees"]]
                )
            ],
            "reminders": {
                "useDefault": False,
                "overrides": [
                    {"method": "email", "minutes": 24 * 60},
                    {"method": "popup", "minutes": 10},
                ],
            },
        }

        event = service.events().insert(calendarId="primary", body=event).execute()
        return pd.DataFrame([event], columns=self.events.get_columns())

    def update_event(self, params: dict = None) -> pd.DataFrame:
        """
        Update event or events in the calendar.
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        df = pd.DataFrame(columns=["eventId", "status"])
        if params["event_id"]:
            start_id = int(params["event_id"])
            end_id = start_id + 1
        elif not params["start_id"]:
            start_id = int(params["end_id"]) - 10
        elif not params["end_id"]:
            end_id = int(params["start_id"]) + 10
        else:
            start_id = int(params["start_id"])
            end_id = int(params["end_id"])

        for i in range(start_id, end_id):
            event = service.events().get(calendarId="primary", eventId=i).execute()
            if params["summary"]:
                event["summary"] = params["summary"]
            if params["location"]:
                event["location"] = params["location"]
            if params["description"]:
                event["description"] = params["description"]
            if params["start"]:
                event["start"]["dateTime"] = params["start"]["dateTime"]
                event["start"]["timeZone"] = params["start"]["timeZone"]
            if params["end"]:
                event["end"]["dateTime"] = params["end"]["dateTime"]
                event["end"]["timeZone"] = params["end"]["timeZone"]
            if params["attendees"]:
                event["attendees"] = [{"email": attendee} for attendee in params["attendees"].split(",")]
            updated_event = service.events().update(calendarId="primary", eventId=event["id"], body=event).execute()
            df = pd.concat(
                [df, pd.DataFrame([{"eventId": updated_event["id"], "status": "updated"}])], ignore_index=True
            )

        return df

    def delete_event(self, params):
        """
        Delete event or events in the calendar.
        Args:
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        if params["event_id"]:
            service.events().delete(calendarId="primary", eventId=params["event_id"]).execute()
            return pd.DataFrame([{"eventId": params["event_id"], "status": "deleted"}])
        else:
            df = pd.DataFrame(columns=["eventId", "status"])
            if not params["start_id"]:
                start_id = int(params["end_id"]) - 10
            elif not params["end_id"]:
                end_id = int(params["start_id"]) + 10
            else:
                start_id = int(params["start_id"])
                end_id = int(params["end_id"])
            for i in range(start_id, end_id):
                service.events().delete(calendarId="primary", eventId=str(i)).execute()
                df = pd.concat([df, pd.DataFrame([{"eventId": str(i), "status": "deleted"}])], ignore_index=True)
            return df

    def call_application_api(self, method_name: str = None, params: dict = None) -> pd.DataFrame:
        """
        Call Google Calendar API and map the data to pandas DataFrame
        Args:
            method_name (str): method name
            params (dict): query parameters
        Returns:
            DataFrame
        """
        if method_name == "get_events":
            return self.get_events(params)
        elif method_name == "create_event":
            return self.create_event(params)
        elif method_name == "update_event":
            return self.update_event(params)
        elif method_name == "delete_event":
            return self.delete_event(params)
        else:
            raise NotImplementedError(f"Unknown method {method_name}")



from typing import Any, Dict

from mindsdb.integrations.handlers.confluence_handler.confluence_api_client import ConfluenceAPIClient
from mindsdb.integrations.handlers.confluence_handler.confluence_tables import (
    ConfluenceBlogPostsTable,
    ConfluenceDatabasesTable,
    ConfluencePagesTable,
    ConfluenceSpacesTable,
    ConfluenceTasksTable,
    ConfluenceWhiteboardsTable,
)
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities import log


logger = log.getLogger(__name__)


class ConfluenceHandler(APIHandler):
    """
    This handler handles the connection and execution of SQL statements on Confluence.
    """

    name = "confluence"

    def __init__(self, name: str, connection_data: Dict, **kwargs: Any) -> None:
        """
        Initializes the handler.

        Args:
            name (str): The name of the handler instance.
            connection_data (Dict): The connection data required to connect to the Confluence API.
            kwargs: Arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False
        self.thread_safe = True

        self._register_table("spaces", ConfluenceSpacesTable(self))
        self._register_table("pages", ConfluencePagesTable(self))
        self._register_table("blogposts", ConfluenceBlogPostsTable(self))
        self._register_table("whiteboards", ConfluenceWhiteboardsTable(self))
        self._register_table("databases", ConfluenceDatabasesTable(self))
        self._register_table("tasks", ConfluenceTasksTable(self))

    def connect(self) -> ConfluenceAPIClient:
        """
        Establishes a connection to the Confluence API.

        Raises:
            ValueError: If the required connection parameters are not provided.

        Returns:
            atlassian.confluence.Confluence: A connection object to the Confluence API.
        """
        if self.is_connected is True:
            return self.connection

        if not all(key in self.connection_data and self.connection_data.get(key) for key in ['api_base', 'username', 'password']):
            raise ValueError('Required parameters (api_base, username, password) must be provided and should not be empty.')

        self.connection = ConfluenceAPIClient(
            url=self.connection_data.get('api_base'),
            username=self.connection_data.get('username'),
            password=self.connection_data.get('password'),
        )

        self.is_connected = True
        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the Confluence API.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)

        try:
            connection = self.connect()
            connection.get_spaces(limit=1)
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to Confluence API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response



"""add_agent_id

Revision ID: f16d4ab03091
Revises: e187961e844a
Create Date: 2023-09-18 10:49:36.290319

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f16d4ab03091'
down_revision = 'e187961e844a'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('chat_bots') as batch_op:
        batch_op.add_column(sa.Column('agent_id', sa.Integer()))
        batch_op.create_foreign_key('fk_agent_id', 'agents', ['agent_id'], ['id'])


def downgrade():
    with op.batch_alter_table('chat_bots') as batch_op:
        batch_op.drop_constraint('fk_agent_id', type_='foreignkey')
        batch_op.drop_column('agent_id')



import unittest
from mindsdb.api.executor.data_types.response_type import RESPONSE_TYPE
from mindsdb.integrations.handlers.rockset_handler.rockset_handler import RocksetHandler


class RocksetHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_args": {
                "host": '127.0.0.1',
                "port": 3306,
                "user": "rockset",
                "password": "rockset"
            }
        }
        cls.handler = RocksetHandler('test_rockset_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.check_connection()

    def test_1_get_tables(self):
        tables = self.handler.get_tables()
        self.assertEqual(tables, [])

    def test_2_get_columns(self):
        columns = self.handler.get_columns('test')
        self.assertEqual(columns, [])

    def test_3_query(self):
        response = self.handler.query('SELECT 1')
        self.assertEqual(response['type'], RESPONSE_TYPE.QUERY)
        self.assertEqual(response['data'], [[1]])



from pathlib import Path

from alembic.command import upgrade, autogen    # noqa
from alembic.config import Config
from alembic.script import ScriptDirectory
from alembic.script.revision import ResolutionError
from alembic.operations import Operations
from alembic.migration import MigrationContext
from alembic import util

import mindsdb.interfaces.storage.db as db
from mindsdb.utilities import log

logger = log.getLogger(__name__)

# This is a migration that is like a 'base version'. Applying only this
# migration to a fresh DB is equivalent to applying all previous migrations.
current_checkpoint = '9f150e4f9a05'


def apply_checkpoint_migration(script) -> None:
    """Apply the checkpoint migration to the database.
    """
    with db.engine.begin() as connection:
        context = MigrationContext.configure(
            connection,
            opts={
                'as_sql': False,
                'starting_rev': None,  # ignore current version
                'destination_rev': current_checkpoint,
            }
        )
        revision = script.get_revision(current_checkpoint)
        if not revision:
            raise util.CommandError(f"Migration {current_checkpoint} not found.")

        op = Operations(context)
        revision.module.upgrade(op)
        context.stamp(script, current_checkpoint)
        connection.commit()


def get_current_revision() -> str | None:
    """ Get the current revision of the database.

    Returns:
        str | None: The current revision of the database.
    """
    with db.engine.begin() as conn:
        mc = MigrationContext.configure(conn)
        return mc.get_current_revision()


def migrate_to_head():
    """ Trying to update database to head revision.
        If alembic unable to recognize current revision (In case when database version is newer than backend)
        then do nothing.
    """

    config_file = Path(__file__).parent / 'alembic.ini'
    config = Config(config_file)

    # mindsdb can runs not from project directory
    script_location_abc = config_file.parent / config.get_main_option('script_location')
    config.set_main_option('script_location', str(script_location_abc))

    script = ScriptDirectory.from_config(config)
    cur_revision = get_current_revision()
    if cur_revision is None:
        apply_checkpoint_migration(script)
        cur_revision = get_current_revision()

    try:
        script.revision_map.get_revision(cur_revision)
    except ResolutionError:
        raise Exception("Database version higher than application.")

    head_rev = script.get_current_head()
    if cur_revision == head_rev:
        logger.debug("The database is in its current state, no updates are required.")
        return

    logger.info("Migrations are available. Applying updates to the database.")
    upgrade(config=config, revision='head')


if __name__ == "__main__":
    # have to import this because
    # all env initialization happens here
    db.init()
    migrate_to_head()



import os
import shutil
import tarfile
import tempfile
import zipfile
from urllib.parse import urlparse

import multipart
import requests
from flask import current_app as ca
from flask import request
from flask_restx import Resource

from mindsdb.api.http.namespaces.configs.files import ns_conf
from mindsdb.api.http.utils import http_error
from mindsdb.metrics.metrics import api_endpoint_metrics
from mindsdb.utilities.config import config
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities import log
from mindsdb.utilities.security import is_private_url, clear_filename, validate_urls
from mindsdb.utilities.fs import safe_extract
from mindsdb.integrations.utilities.files.file_reader import FileProcessingError

logger = log.getLogger(__name__)
MAX_FILE_SIZE = 1024 * 1024 * 100  # 100Mb


@ns_conf.route("/")
class FilesList(Resource):
    @ns_conf.doc("get_files_list")
    @api_endpoint_metrics("GET", "/files")
    def get(self):
        """List all files"""
        return ca.file_controller.get_files()


@ns_conf.route("/<name>")
@ns_conf.param("name", "MindsDB's name for file")
class File(Resource):
    @ns_conf.doc("put_file")
    @api_endpoint_metrics("PUT", "/files/file")
    def put(self, name: str):
        """add new file
        params in FormData:
            - file
            - original_file_name [optional]
        """

        data = {}
        mindsdb_file_name = name.lower()

        existing_file_names = ca.file_controller.get_files_names()

        def on_field(field):
            name = field.field_name.decode()
            value = field.value.decode()
            data[name] = value

        file_object = None

        def on_file(file):
            nonlocal file_object
            data["file"] = file.file_name.decode()
            file_object = file.file_object

        temp_dir_path = tempfile.mkdtemp(prefix="mindsdb_file_")

        if request.headers["Content-Type"].startswith("multipart/form-data"):
            parser = multipart.create_form_parser(
                headers=request.headers,
                on_field=on_field,
                on_file=on_file,
                config={
                    "UPLOAD_DIR": temp_dir_path.encode(),  # bytes required
                    "UPLOAD_KEEP_FILENAME": True,
                    "UPLOAD_KEEP_EXTENSIONS": True,
                    "MAX_MEMORY_FILE_SIZE": 0,
                },
            )

            while True:
                chunk = request.stream.read(8192)
                if not chunk:
                    break
                parser.write(chunk)
            parser.finalize()
            parser.close()

            if file_object is not None:
                if not file_object.closed:
                    try:
                        file_object.flush()
                    except (AttributeError, ValueError, OSError):
                        logger.debug("Failed to flush file_object before closing.", exc_info=True)
                    file_object.close()
                file_object = None
        else:
            data = request.json

        if mindsdb_file_name in existing_file_names:
            return http_error(
                400,
                "File already exists",
                f"File with name '{data['file']}' already exists",
            )

        if data.get("source_type") == "url":
            url = data["source"]
            try:
                url = urlparse(url)
                if not (url.scheme and url.netloc):
                    raise ValueError()
                url = url.geturl()
            except Exception:
                return http_error(
                    400,
                    "Invalid URL",
                    f"The URL is not valid: {data['source']}",
                )

            url_file_upload_enabled = config["url_file_upload"]["enabled"]
            if url_file_upload_enabled is False:
                return http_error(400, "URL file upload is disabled.", "URL file upload is disabled.")

            allowed_origins = config["url_file_upload"]["allowed_origins"]
            disallowed_origins = config["url_file_upload"]["disallowed_origins"]

            if validate_urls(url, allowed_origins, disallowed_origins) is False:
                return http_error(
                    400,
                    "Invalid URL",
                    "URL is not allowed for security reasons. Allowed hosts are: "
                    f"{', '.join(allowed_origins) if allowed_origins else 'not specified'}.",
                )

            data["file"] = clear_filename(data["name"])
            is_cloud = config.get("cloud", False)
            if is_cloud:
                if is_private_url(url):
                    return http_error(400, f"URL is private: {url}")

                if ctx.user_class != 1:
                    info = requests.head(url, timeout=30)
                    file_size = info.headers.get("Content-Length")
                    try:
                        file_size = int(file_size)
                    except Exception:
                        pass

                    if file_size is None:
                        return http_error(
                            400,
                            "Error getting file info",
                            "Сan't determine remote file size",
                        )
                    if file_size > MAX_FILE_SIZE:
                        return http_error(400, "File is too big", f"Upload limit for file is {MAX_FILE_SIZE >> 20} MB")
            with requests.get(url, stream=True) as r:
                if r.status_code != 200:
                    return http_error(400, "Error getting file", f"Got status code: {r.status_code}")
                file_path = os.path.join(temp_dir_path, data["file"])
                with open(file_path, "wb") as f:
                    for chunk in r.iter_content(chunk_size=8192):
                        f.write(chunk)

        original_file_name = clear_filename(data.get("original_file_name"))

        file_path = os.path.join(temp_dir_path, data["file"])
        lp = file_path.lower()
        if lp.endswith((".zip", ".tar.gz")):
            if lp.endswith(".zip"):
                with zipfile.ZipFile(file_path) as f:
                    f.extractall(temp_dir_path)
            elif lp.endswith(".tar.gz"):
                with tarfile.open(file_path) as f:
                    safe_extract(f, temp_dir_path)
            os.remove(file_path)
            files = os.listdir(temp_dir_path)
            if len(files) != 1:
                os.rmdir(temp_dir_path)
                return http_error(400, "Wrong content.", "Archive must contain only one data file.")
            file_path = os.path.join(temp_dir_path, files[0])
            mindsdb_file_name = files[0]
            if not os.path.isfile(file_path):
                os.rmdir(temp_dir_path)
                return http_error(400, "Wrong content.", "Archive must contain data file in root.")

        try:
            ca.file_controller.save_file(mindsdb_file_name, file_path, file_name=original_file_name)
        except FileProcessingError as e:
            return http_error(400, "Error", str(e))
        except Exception as e:
            return http_error(500, "Error", str(e))
        finally:
            shutil.rmtree(temp_dir_path, ignore_errors=True)

        return "", 200

    @ns_conf.doc("delete_file")
    @api_endpoint_metrics("DELETE", "/files/file")
    def delete(self, name: str):
        """delete file"""

        try:
            ca.file_controller.delete_file(name)
        except Exception as e:
            logger.error(e)
            return http_error(
                400,
                "Error deleting file",
                f"There was an error while tring to delete file with name '{name}'",
            )
        return "", 200



import pandas as pd
import json
import symbl
from typing import List
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.handlers.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.utilities import log
from mindsdb_sql_parser import ast

logger = log.getLogger(__name__)


class GetConversationTable(APITable):
    """The Get Conversation Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Get the conversation Id for the given audio file" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            conversation id

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'get_conversation_id',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'audio_url':
                if op == '=':
                    search_params["audio_url"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for audio_url column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("audio_url" in search_params)

        if not filter_flag:
            raise NotImplementedError("audio_url column has to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        payload = {"url": search_params.get("audio_url")}
        conversation_object = symbl.Audio.process_url(payload=payload, credentials=self.handler.credentials)

        df = pd.json_normalize({"conversation_id": conversation_object.get_conversation_id()})

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "conversation_id"
        ]


class GetMessagesTable(APITable):
    """The Get Messages Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Get the messages for the given conversation id" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Messages

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'get_messages',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'conversation_id':
                if op == '=':
                    search_params["conversation_id"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for conversation_id column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("conversation_id" in search_params)

        if not filter_flag:
            raise NotImplementedError("conversation_id column has to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        resp = symbl.Conversations.get_messages(conversation_id=search_params.get("conversation_id"), credentials=self.handler.credentials)

        resp = self.parse_response(resp)

        df = pd.json_normalize(resp["messages"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def parse_response(self, res):
        return json.loads(json.dumps(res.to_dict(), default=str))

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "conversation_id",
            "end_time",
            "id",
            "phrases",
            "sentiment",
            "start_time",
            "text",
            "words",
            "_from.email",
            "_from.id",
            "_from.name"
        ]


class GetTopicsTable(APITable):
    """The Get Topics Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Get the topics for the given conversation id" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Topics

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'get_topics',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'conversation_id':
                if op == '=':
                    search_params["conversation_id"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for conversation_id column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("conversation_id" in search_params)

        if not filter_flag:
            raise NotImplementedError("conversation_id column has to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        resp = symbl.Conversations.get_topics(conversation_id=search_params.get("conversation_id"), credentials=self.handler.credentials)

        resp = self.parse_response(resp)

        df = pd.json_normalize(resp["topics"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def parse_response(self, res):
        return json.loads(json.dumps(res.to_dict(), default=str))

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "id",
            "text",
            "type",
            "score",
            "message_ids",
            "entities",
            "sentiment",
            "parent_refs"
        ]


class GetQuestionsTable(APITable):
    """The Get Questions Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Get the questions for the given conversation id" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Questions

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'get_questions',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'conversation_id':
                if op == '=':
                    search_params["conversation_id"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for conversation_id column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("conversation_id" in search_params)

        if not filter_flag:
            raise NotImplementedError("conversation_id column has to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        resp = symbl.Conversations.get_questions(conversation_id=search_params.get("conversation_id"), credentials=self.handler.credentials)

        resp = self.parse_response(resp)

        df = pd.json_normalize(resp["questions"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def parse_response(self, res):
        return json.loads(json.dumps(res.to_dict(), default=str))

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "id",
            "text",
            "type",
            "score",
            "message_ids",
            "_from.id",
            "_from.name",
            "_from.user_id"
        ]


class GetFollowUpsTable(APITable):
    """The Get FollowUps Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Get the follow ups for the given conversation id" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            follow up Questions

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'get_follow_ups',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'conversation_id':
                if op == '=':
                    search_params["conversation_id"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for conversation_id column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("conversation_id" in search_params)

        if not filter_flag:
            raise NotImplementedError("conversation_id column has to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        resp = symbl.Conversations.get_follow_ups(conversation_id=search_params.get("conversation_id"), credentials=self.handler.credentials)

        resp = self.parse_response(resp)

        df = pd.json_normalize(resp["follow_ups"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def parse_response(self, res):
        return json.loads(json.dumps(res.to_dict(), default=str))

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "id",
            "text",
            "type",
            "score",
            "message_ids",
            "entities",
            "phrases",
            "definitive",
            "due_by",
            "_from.id",
            "_from.name",
            "_from.user_id",
            "assignee.id",
            "assignee.name",
            "assignee.email"
        ]


class GetActionItemsTable(APITable):
    """The Get Action items Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Get the action items for the given conversation id" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            action items

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'get_action_items',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'conversation_id':
                if op == '=':
                    search_params["conversation_id"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for conversation_id column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("conversation_id" in search_params)

        if not filter_flag:
            raise NotImplementedError("conversation_id column has to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        resp = symbl.Conversations.get_action_items(conversation_id=search_params.get("conversation_id"), credentials=self.handler.credentials)

        resp = self.parse_response(resp)

        df = pd.json_normalize(resp["action_items"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def parse_response(self, res):
        return json.loads(json.dumps(res.to_dict(), default=str))

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "id",
            "text",
            "type",
            "score",
            "message_ids",
            "entities",
            "phrases",
            "definitive",
            "due_by",
            "_from.id",
            "_from.name",
            "_from.user_id",
            "assignee.id",
            "assignee.name",
            "assignee.email"
        ]


class GetAnalyticsTable(APITable):
    """The Get Analytics Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Get the analytics for the given conversation id" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            metrics

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'get_analytics',
            self.get_columns()
        )

        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        subset_where_conditions = []
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'conversation_id':
                if op == '=':
                    search_params["conversation_id"] = arg2
                else:
                    raise NotImplementedError("Only '=' operator is supported for conversation_id column.")

            elif arg1 in self.get_columns():
                subset_where_conditions.append([op, arg1, arg2])

        filter_flag = ("conversation_id" in search_params)

        if not filter_flag:
            raise NotImplementedError("conversation_id column has to be present in where clause.")

        df = pd.DataFrame(columns=self.get_columns())

        resp = symbl.Conversations.get_analytics(conversation_id=search_params.get("conversation_id"), credentials=self.handler.credentials)

        resp = self.parse_response(resp)

        df = pd.json_normalize(resp["metrics"])

        select_statement_executor = SELECTQueryExecutor(
            df,
            selected_columns,
            subset_where_conditions,
            order_by_conditions,
            result_limit
        )

        df = select_statement_executor.execute_query()

        return df

    def parse_response(self, res):
        return json.loads(json.dumps(res.to_dict(), default=str))

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """

        return [
            "type",
            "percent",
            "seconds"
        ]



import os
import time
import tempfile
import threading
from pathlib import Path
from typing import Optional, List, Tuple

import psutil

from mindsdb.utilities import log

logger = log.getLogger(__name__)


def get_tmp_dir() -> Path:
    return Path(tempfile.gettempdir()).joinpath("mindsdb")


def _get_process_mark_id(unified: bool = False) -> str:
    """Creates a text that can be used to identify process+thread
    Args:
        unified: bool, if True then result will be same for same process+thread
    Returns:
        mark of process+thread
    """
    mark = f"{os.getpid()}-{threading.get_native_id()}"
    if unified is True:
        return mark
    return f"{mark}-{str(time.time()).replace('.', '')}"


def create_process_mark(folder="learn"):
    p = get_tmp_dir().joinpath(f"processes/{folder}/")
    p.mkdir(parents=True, exist_ok=True)
    mark = _get_process_mark_id()
    p.joinpath(mark).touch()
    return mark


def set_process_mark(folder: str, mark: str) -> None:
    """touch new file which will be process mark

    Args:
        folder (str): where create the file
        mark (str): file name

    Returns:
        str: process mark
    """
    p = get_tmp_dir().joinpath(f"processes/{folder}/")
    p.mkdir(parents=True, exist_ok=True)
    mark = f"{os.getpid()}-{threading.get_native_id()}-{mark}"
    p.joinpath(mark).touch()
    return mark


def delete_process_mark(folder: str = "learn", mark: Optional[str] = None):
    if mark is None:
        mark = _get_process_mark_id()
    p = get_tmp_dir().joinpath(f"processes/{folder}/").joinpath(mark)
    if p.exists():
        p.unlink()


def clean_process_marks():
    """delete all existing processes marks"""
    logger.debug("Deleting PIDs..")
    p = get_tmp_dir().joinpath("processes/")
    if p.exists() is False:
        return
    for path in p.iterdir():
        if path.is_dir() is False:
            return
        for file in path.iterdir():
            file.unlink()


def get_processes_dir_files_generator() -> Tuple[Path, int, int]:
    """Get files from processes dir

    Yields:
        Tuple[Path, int, int]: file object, process is and thread id
    """
    p = get_tmp_dir().joinpath("processes/")
    if p.exists() is False:
        return
    for path in p.iterdir():
        if path.is_dir() is False:
            continue
        for file in path.iterdir():
            parts = file.name.split("-")
            process_id = int(parts[0])
            thread_id = int(parts[1])
            yield file, process_id, thread_id


def clean_unlinked_process_marks() -> List[int]:
    """delete marks that does not have corresponded processes/threads

    Returns:
        List[int]: list with ids of unexisting processes
    """
    deleted_pids = []

    for file, process_id, thread_id in get_processes_dir_files_generator():
        try:
            process = psutil.Process(process_id)
            if process.status() in (psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD):
                raise psutil.NoSuchProcess(process_id)

            threads = process.threads()
            try:
                next(t for t in threads if t.id == thread_id)
            except StopIteration:
                logger.warning(f"We have mark for process/thread {process_id}/{thread_id} but it does not exists")
                deleted_pids.append(process_id)
                file.unlink()

        except psutil.AccessDenied:
            logger.warning(f"access to {process_id} denied")

            continue

        except psutil.NoSuchProcess:
            logger.warning(f"We have mark for process/thread {process_id}/{thread_id} but it does not exists")
            deleted_pids.append(process_id)
            file.unlink()
    return deleted_pids


def create_pid_file():
    """
    Create mindsdb process pid file. Check if previous process exists and is running
    """

    if os.environ.get("USE_PIDFILE") != "1":
        return

    p = get_tmp_dir()
    p.mkdir(parents=True, exist_ok=True)
    pid_file = p.joinpath("pid")
    if pid_file.exists():
        # if process exists raise exception
        pid = pid_file.read_text().strip()
        try:
            psutil.Process(int(pid))
            raise Exception(f"Found PID file with existing process: {pid} {pid_file}")
        except (psutil.Error, ValueError):
            ...

        logger.warning(f"Found existing PID file {pid_file}({pid}), removing")
        pid_file.unlink()

    pid_file.write_text(str(os.getpid()))


def delete_pid_file():
    """
    Remove existing process pid file if it matches current process
    """

    if os.environ.get("USE_PIDFILE") != "1":
        return

    pid_file = get_tmp_dir().joinpath("pid")

    if not pid_file.exists():
        return

    pid = pid_file.read_text().strip()
    if pid != str(os.getpid()):
        logger.warning(f"Process id in PID file ({pid_file}) doesn't match mindsdb pid")
        return

    pid_file.unlink()


def __is_within_directory(directory, target):
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)
    prefix = os.path.commonprefix([abs_directory, abs_target])
    return prefix == abs_directory


def safe_extract(tarfile, path=".", members=None, *, numeric_owner=False):
    # for py >= 3.12
    if hasattr(tarfile, "data_filter"):
        tarfile.extractall(path, members=members, numeric_owner=numeric_owner, filter="data")
        return

    # for py < 3.12
    for member in tarfile.getmembers():
        member_path = os.path.join(path, member.name)
        if not __is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")
    tarfile.extractall(path, members=members, numeric_owner=numeric_owner)



from mindsdb.integrations.handlers.mysql_handler import Handler as MySQLHandler


class PlanetScaleHandler(MySQLHandler):
    """
    This handler handles the connection and execution of queries against PlanetScale.
    """
    name = 'planet_scale'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)



from pydantic import BaseModel, AnyUrl, model_validator
from urllib.parse import urlparse


class ConnectionConfig(BaseModel):
    # TODO: For now validate AnyURL since MySQLDsn wasn't working
    url: AnyUrl = None
    host: str = None
    port: int = 3306
    user: str = None
    password: str = None
    database: str = None

    @model_validator(mode="before")
    def check_db_params(cls, values):
        """Ensures either URL is provided or all individual parameters are provided."""
        url = values.get("url")
        host = values.get("host")
        user = values.get("user")
        password = values.get("password")
        database = values.get("database")
        if not url and not (host and user and password and database):
            raise ValueError(
                "Either a valid URL or required parameters (host, user, password, database) must be provided."
            )

        if url:
            parsed = urlparse(url)
            values["host"] = parsed.hostname or host
            values["port"] = parsed.port if parsed.port is not None else 3306
            values["user"] = parsed.username or user
            values["password"] = parsed.password or password
            values["database"] = parsed.path[1:] if parsed.path else database

            # mysql connector raise error if url is provided
            values.pop("url", None)

            return values

        if not url:
            for param in ["host", "user", "password", "database"]:
                if not values.get(param):
                    raise ValueError(f"{param} is required when URL is not provided.")
        return values

    class Config:
        str_min_length = 1
        str_strip_whitespace = True



import json
from typing import Any, Dict, Optional, Text

from google.cloud.bigquery import Client, QueryJobConfig, DEFAULT_RETRY
from google.api_core.exceptions import BadRequest, NotFound
import pandas as pd
from sqlalchemy_bigquery.base import BigQueryDialect

from mindsdb.utilities import log
from mindsdb_sql_parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import MetaDatabaseHandler
from mindsdb.utilities.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.utilities.handlers.auth_utilities.google import GoogleServiceAccountOAuth2Manager
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)

logger = log.getLogger(__name__)


class BigQueryHandler(MetaDatabaseHandler):
    """
    This handler handles connection and execution of Google BigQuery statements.
    """

    name = "bigquery"

    def __init__(self, name: Text, connection_data: Dict, **kwargs: Any):
        super().__init__(name)
        self.connection_data = connection_data
        self.client = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self):
        """
        Establishes a connection to a BigQuery warehouse.

        Raises:
            ValueError: If the required connection parameters are not provided or if the credentials cannot be parsed.
            mindsdb.integrations.utilities.handlers.auth_utilities.exceptions.NoCredentialsException: If none of the required forms of credentials are provided.
            mindsdb.integrations.utilities.handlers.auth_utilities.exceptions.AuthException: If authentication fails.

        Returns:
            google.cloud.bigquery.client.Client: The client object for the BigQuery connection.
        """
        if self.is_connected is True:
            return self.connection

        # Mandatory connection parameters
        if not all(key in self.connection_data for key in ["project_id", "dataset"]):
            raise ValueError("Required parameters (project_id, dataset) must be provided.")

        service_account_json = self.connection_data.get("service_account_json")
        if isinstance(service_account_json, str):
            # GUI send it as str
            try:
                service_account_json = json.loads(service_account_json)
            except json.decoder.JSONDecodeError:
                raise ValueError("'service_account_json' is not valid JSON")
        if isinstance(service_account_json, dict) and isinstance(service_account_json.get("private_key"), str):
            # some editors may escape new line symbol, also replace windows-like newlines
            service_account_json["private_key"] = (
                service_account_json["private_key"].replace("\\n", "\n").replace("\r\n", "\n")
            )

        google_sa_oauth2_manager = GoogleServiceAccountOAuth2Manager(
            credentials_file=self.connection_data.get("service_account_keys"),
            credentials_json=service_account_json,
        )
        credentials = google_sa_oauth2_manager.get_oauth2_credentials()

        client = Client(project=self.connection_data["project_id"], credentials=credentials)
        self.is_connected = True
        self.connection = client
        return self.connection

    def disconnect(self):
        """
        Closes the connection to the BigQuery warehouse if it's currently open.
        """
        if self.is_connected is False:
            return
        self.connection.close()
        self.is_connected = False

    def check_connection(self) -> StatusResponse:
        """
        Checks the status of the connection to the BigQuery warehouse.

        Returns:
            StatusResponse: An object containing the success status and an error message if an error occurs.
        """
        response = StatusResponse(False)

        try:
            connection = self.connect()
            connection.query("SELECT 1;", timeout=10, retry=DEFAULT_RETRY.with_deadline(10))

            # Check if the dataset exists
            connection.get_dataset(self.connection_data["dataset"])

            response.success = True
        except (BadRequest, ValueError) as e:
            logger.error(f"Error connecting to BigQuery {self.connection_data['project_id']}, {e}!")
            response.error_message = e
        except NotFound:
            response.error_message = (
                f"Error connecting to BigQuery {self.connection_data['project_id']}: "
                f"dataset '{self.connection_data['dataset']}' not found"
            )

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> Response:
        """
        Executes a SQL query on the BigQuery warehouse and returns the result.

        Args:
            query (str): The SQL query to be executed.

        Returns:
            Response: A response object containing the result of the query or an error message.
        """
        connection = self.connect()
        try:
            job_config = QueryJobConfig(
                default_dataset=f"{self.connection_data['project_id']}.{self.connection_data['dataset']}"
            )
            query = connection.query(query, job_config=job_config)
            result = query.to_dataframe()
            if not result.empty:
                response = Response(RESPONSE_TYPE.TABLE, result)
            else:
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            logger.error(f"Error running query: {query} on {self.connection_data['project_id']}!")
            response = Response(RESPONSE_TYPE.ERROR, error_message=str(e))
        return response

    def query(self, query: ASTNode) -> Response:
        """
        Executes a SQL query represented by an ASTNode and retrieves the data.

        Args:
            query (ASTNode): An ASTNode representing the SQL query to be executed.

        Returns:
            Response: The response from the `native_query` method, containing the result of the SQL query execution.
        """
        renderer = SqlalchemyRender(BigQueryDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Retrieves a list of all non-system tables and views in the configured dataset of the BigQuery warehouse.

        Returns:
            Response: A response object containing the list of tables and views, formatted as per the `Response` class.
        """
        query = f"""
            SELECT table_name, table_schema, table_type
            FROM `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.INFORMATION_SCHEMA.TABLES`
            WHERE table_type IN ('BASE TABLE', 'VIEW')
        """
        result = self.native_query(query)
        return result

    def get_columns(self, table_name) -> Response:
        """
        Retrieves column details for a specified table in the configured dataset of the BigQuery warehouse.

        Args:
            table_name (str): The name of the table for which to retrieve column information.

        Returns:
            Response: A response object containing the column details, formatted as per the `Response` class.
        Raises:
            ValueError: If the 'table_name' is not a valid string.
        """
        query = f"""
            SELECT column_name AS Field, data_type as Type
            FROM `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.INFORMATION_SCHEMA.COLUMNS`
            WHERE table_name = '{table_name}'
        """
        result = self.native_query(query)
        return result

    def meta_get_tables(self, table_names: Optional[list] = None) -> Response:
        """
        Retrieves table metadata for the specified tables (or all tables if no list is provided).

        Args:
            table_names (list): A list of table names for which to retrieve metadata information.

        Returns:
            Response: A response object containing the metadata information, formatted as per the `Response` class.
        """
        query = f"""
            SELECT
                t.table_name,
                t.table_schema,
                t.table_type,
                st.row_count
            FROM 
                `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.INFORMATION_SCHEMA.TABLES` AS t
            JOIN 
                `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.__TABLES__` AS st
            ON 
                t.table_name = st.table_id
            WHERE 
                t.table_type IN ('BASE TABLE', 'VIEW')
        """

        if table_names is not None and len(table_names) > 0:
            table_names = [f"'{t}'" for t in table_names]
            query += f" AND t.table_name IN ({','.join(table_names)})"

        result = self.native_query(query)
        return result

    def meta_get_columns(self, table_names: Optional[list] = None) -> Response:
        """
        Retrieves column metadata for the specified tables (or all tables if no list is provided).

        Args:
            table_names (list): A list of table names for which to retrieve column metadata.

        Returns:
            Response: A response object containing the column metadata.
        """
        query = f"""
            SELECT 
                table_name,
                column_name,
                data_type,
                column_default,
                CASE is_nullable
                    WHEN 'YES' THEN TRUE
                    ELSE FALSE
                END AS is_nullable
            FROM 
                `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.INFORMATION_SCHEMA.COLUMNS`
        """

        if table_names is not None and len(table_names) > 0:
            table_names = [f"'{t}'" for t in table_names]
            query += f" WHERE table_name IN ({','.join(table_names)})"

        result = self.native_query(query)
        return result

    def meta_get_column_statistics_for_table(self, table_name: str, columns: list) -> Response:
        """
        Retrieves statistics for the specified columns in a table.

        Args:
            table_name (str): The name of the table.
            columns (list): A list of column names to retrieve statistics for.

        Returns:
            Response: A response object containing the column statistics.
        """
        # To avoid hitting BigQuery's query size limits, we will chunk the columns into batches.
        # This is because the queries are combined using UNION ALL, which can lead to very large queries if there are many columns.
        BATCH_SIZE = 20

        def chunked(lst, n):
            """
            Yields successive n-sized chunks from lst.
            """
            for i in range(0, len(lst), n):
                yield lst[i : i + n]

        queries = []
        for column_batch in chunked(columns, BATCH_SIZE):
            batch_queries = []
            for column in column_batch:
                batch_queries.append(
                    f"""
                    SELECT
                        '{table_name}' AS table_name,
                        '{column}' AS column_name,
                        SAFE_DIVIDE(COUNTIF({column} IS NULL), COUNT(*)) * 100 AS null_percentage,
                        CAST(MIN(`{column}`) AS STRING) AS minimum_value,
                        CAST(MAX(`{column}`) AS STRING) AS maximum_value,
                        COUNT(DISTINCT {column}) AS distinct_values_count
                    FROM
                        `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.{table_name}`
                    """
                )

            query = " UNION ALL ".join(batch_queries)
            queries.append(query)

        results = []
        for query in queries:
            try:
                result = self.native_query(query)
                if result.resp_type == RESPONSE_TYPE.TABLE:
                    results.append(result.data_frame)
                else:
                    logger.error(f"Error retrieving column statistics for table {table_name}: {result.error_message}")
            except Exception as e:
                logger.error(f"Exception occurred while retrieving column statistics for table {table_name}: {e}")

        if not results:
            logger.warning(f"No column statistics could be retrieved for table {table_name}.")
            return Response(
                RESPONSE_TYPE.ERROR, error_message=f"No column statistics could be retrieved for table {table_name}."
            )
        return Response(RESPONSE_TYPE.TABLE, pd.concat(results, ignore_index=True) if results else pd.DataFrame())

    def meta_get_primary_keys(self, table_names: Optional[list] = None) -> Response:
        """
        Retrieves primary key information for the specified tables (or all tables if no list is provided).

        Args:
            table_names (list): A list of table names for which to retrieve primary key information.

        Returns:
            Response: A response object containing the primary key information.
        """
        query = f"""
            SELECT
                tc.table_name,
                kcu.column_name,
                kcu.ordinal_position,
                tc.constraint_name,
            FROM
                `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.INFORMATION_SCHEMA.TABLE_CONSTRAINTS` AS tc
            JOIN
                `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.INFORMATION_SCHEMA.KEY_COLUMN_USAGE` AS kcu
            ON
                tc.constraint_name = kcu.constraint_name
            WHERE
                tc.constraint_type = 'PRIMARY KEY'
        """

        if table_names is not None and len(table_names) > 0:
            table_names = [f"'{t}'" for t in table_names]
            query += f" AND tc.table_name IN ({','.join(table_names)})"

        result = self.native_query(query)
        return result

    def meta_get_foreign_keys(self, table_names: Optional[list] = None) -> Response:
        """
        Retrieves foreign key information for the specified tables (or all tables if no list is provided).

        Args:
            table_names (list): A list of table names for which to retrieve foreign key information.

        Returns:
            Response: A response object containing the foreign key information.
        """
        query = f"""
            SELECT
                ccu.table_name AS parent_table_name,
                ccu.column_name AS parent_column_name,
                kcu.table_name AS child_table_name,
                kcu.column_name AS child_column_name,
                tc.constraint_name
            FROM
                `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.INFORMATION_SCHEMA.TABLE_CONSTRAINTS` AS tc
            JOIN
                `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.INFORMATION_SCHEMA.KEY_COLUMN_USAGE` AS kcu
            ON
                tc.constraint_name = kcu.constraint_name
            JOIN
                `{self.connection_data["project_id"]}.{self.connection_data["dataset"]}.INFORMATION_SCHEMA.CONSTRAINT_COLUMN_USAGE` AS ccu
            ON
                tc.constraint_name = ccu.constraint_name
            WHERE
                tc.constraint_type = 'FOREIGN KEY'
        """

        if table_names is not None and len(table_names) > 0:
            table_names = [f"'{t}'" for t in table_names]
            query += f" AND tc.table_name IN ({','.join(table_names)})"

        result = self.native_query(query)
        return result



import ast as py_ast

import dateutil.parser
from bson import ObjectId

from .mongodb_query import MongoQuery


class MongodbParser:
    """
    Converts string into MongoQuery
    """

    def from_string(self, call_str):
        tree = py_ast.parse(call_str.strip(), mode="eval")
        calls = self.process(tree.body)
        # first call contents collection
        method1 = calls[0]["method"]
        if len(method1) < 2:
            raise IndexError("Collection not found")
        collection = method1[-2]

        mquery = MongoQuery(collection)

        # keep only last name
        calls[0]["method"] = [method1[-1]]

        # convert method names: get first item of list
        for c in calls:
            mquery.add_step({"method": c["method"][0], "args": c["args"]})

        return mquery

    def process(self, node):
        if isinstance(node, py_ast.Call):
            previous_call = None

            args = []
            for node2 in node.args:
                args.append(self.process(node2))

            # check functions
            if isinstance(node.func, py_ast.Name):
                # it is just name
                func = node.func.id

                # special functions:
                if func == "ISODate":
                    return dateutil.parser.isoparse(args[0])
                if func == "ObjectId":
                    return ObjectId(args[0])
            elif isinstance(node.func, py_ast.Attribute):
                # it can be an attribute or pipeline
                previous_call, func = self.process_func_name(node.func)
            else:
                raise NotImplementedError(f"Unknown function type: {node.func}")

            call = [{"method": func, "args": args}]
            if previous_call is not None:
                call = previous_call + call

            return call

        if isinstance(node, py_ast.List):
            elements = []
            for node2 in node.elts:
                elements.append(self.process(node2))
            return elements

        if isinstance(node, py_ast.Dict):
            keys = []
            for node2 in node.keys:
                if isinstance(node2, py_ast.Constant):
                    value = node2.value
                elif isinstance(node2, py_ast.Str):  # py37
                    value = node2.s
                elif isinstance(node2, py_ast.Name):
                    value = node2.id
                else:
                    raise NotImplementedError(f"Unknown dict key {node2}")

                keys.append(value)

            values = []
            for node2 in node.values:
                values.append(self.process(node2))

            return dict(zip(keys, values))

        if isinstance(node, py_ast.Name):
            # special attributes
            name = node.id
            if name == "true":
                return True
            elif name == "false":
                return False
            elif name == "null":
                return None

        if isinstance(node, py_ast.Constant):
            return node.value

        # ---- python 3.7 objects -----
        if isinstance(node, py_ast.Str):
            return node.s

        if isinstance(node, py_ast.Num):
            return node.n

        # -----------------------------

        if isinstance(node, py_ast.UnaryOp):
            if isinstance(node.op, py_ast.USub):
                value = self.process(node.operand)
                return -value

        raise NotImplementedError(f"Unknown node {node}")

    def process_func_name(self, node):
        previous_call = None
        if isinstance(node, py_ast.Attribute):
            attribute = node
            # multilevel attribute

            obj_name = []
            while isinstance(attribute, py_ast.Attribute):
                obj_name.insert(0, attribute.attr)
                attribute = attribute.value

            if isinstance(attribute, py_ast.Name):
                obj_name.insert(0, attribute.id)

            if isinstance(attribute, py_ast.Call):
                # is pipeline
                previous_call = self.process(attribute)

            return previous_call, obj_name



import os
import shutil
import tempfile
from io import BytesIO, StringIO
from pathlib import Path

import pandas
import pytest
from mindsdb_sql_parser.exceptions import ParsingException
from mindsdb_sql_parser.ast import (
    CreateTable,
    DropTables,
    Identifier,
    Insert,
    TableColumn,
    Update,
)

from mindsdb.integrations.handlers.file_handler.file_handler import FileHandler
from mindsdb.integrations.libs.response import RESPONSE_TYPE

from mindsdb.integrations.utilities.files.file_reader import (
    FileReader,
    FileProcessingError,
)


# Define a table to use as content for all of the file types
# This data needs to match that saved in the files in the ./data/ dir (except pdf and txt files)
test_file_content = [
    ["col_one", "col_two", "col_three", "col_four"],
    [1, -1, 0.1, "A"],
    [2, -2, 0.2, "B"],
    [3, -3, 0.3, "C"],
]

test_excel_sheet_content = [
    ["Sheet_Name"],
    ["Sheet1"],
]

file_records = [("one", 1, test_file_content[0]), ("two", 2, test_file_content[0])]


class MockFileController:
    """
    Pretends to be a file controller. Gives details of 'file_records' above, and mocks file deletion.
    We're not testing the file controller here, so we don't need to rely on it.
    """

    def get_files(self):
        return [
            {
                "name": record[0],
                "row_count": record[1],
                "columns": record[2],
            }
            for record in file_records
        ]

    def get_files_names(self):
        return [file["name"] for file in self.get_files()]

    def get_file_path(self, name):
        return True

    def get_file_meta(self, *args, **kwargs):
        return self.get_files()[0]

    def delete_file(self, name):
        return True

    def save_file(self, name, file_path, file_name=None):
        return True

    def get_file_data(self, name, page_name=None):
        return pandas.DataFrame(test_file_content[1:], columns=test_file_content[0])

    def set_file_data(self, name, df, page_name=None):
        return True


def curr_dir():
    return os.path.dirname(os.path.realpath(__file__))


def csv_file() -> str:
    return os.path.join(curr_dir(), "data", "test.csv")


def xlsx_file() -> str:
    return os.path.join(curr_dir(), "data", "test.xlsx")


def json_file() -> str:
    return os.path.join(curr_dir(), "data", "test.json")


def parquet_file() -> str:
    return os.path.join(curr_dir(), "data", "test.parquet")


def pdf_file() -> str:
    return os.path.join(curr_dir(), "data", "test.pdf")


def txt_file() -> str:
    return os.path.join(curr_dir(), "data", "test.txt")


class TestIsItX:
    """Tests all of the 'is_it_x()' functions to determine a file's type"""

    def test_is_it_csv(self):
        # We can't test xlsx or parquet here because they're binary files
        for file_path, result in ((csv_file(), True), (json_file(), False)):
            with open(file_path, "r") as fh:
                assert FileReader.is_csv(StringIO(fh.read())) is result

    def test_format(self):
        for file_path, result in (
            (csv_file(), "csv"),
            (xlsx_file(), "xlsx"),
            (json_file(), "json"),
            (parquet_file(), "parquet"),
            (txt_file(), "txt"),
            (pdf_file(), "pdf"),
        ):
            assert FileReader(path=file_path).get_format() == result

    def test_is_it_json(self):
        # We can't test xlsx or parquet here because they're binary files
        for file_path, result in (
            (csv_file(), False),
            (json_file(), True),
            (txt_file(), False),
        ):
            with open(file_path, "r") as fh:
                assert FileReader.is_json(StringIO(fh.read())) is result

    def test_is_it_parquet(self):
        for file_path, result in (
            (csv_file(), False),
            (xlsx_file(), False),
            (json_file(), False),
            (parquet_file(), True),
            (txt_file(), False),
            (pdf_file(), False),
        ):
            with open(file_path, "rb") as fh:
                assert FileReader.is_parquet(BytesIO(fh.read())) is result


class TestQuery:
    """Tests all of the scenarios relating to the query() function"""

    def test_query_drop(self, monkeypatch):
        """Test a valid drop table query"""

        def mock_delete(self, name):
            return True

        monkeypatch.setattr(MockFileController, "delete_file", mock_delete)
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.query(DropTables([Identifier(parts=["one"])]))

        assert response.type == RESPONSE_TYPE.OK

    def test_query_drop_bad_delete(self, monkeypatch):
        """Test an invalid drop table query"""

        def mock_delete(self, name):
            raise Exception("File delete error")

        monkeypatch.setattr(MockFileController, "delete_file", mock_delete)
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.query(DropTables([Identifier(parts=["one"])]))

        assert response.type == RESPONSE_TYPE.ERROR

    def test_query_insert(self, monkeypatch):
        """Test an invalid insert query"""
        # Create a temporary file to save the csv file to.
        csv_file_path = csv_file()
        csv_tmp = os.path.join(tempfile.gettempdir(), "test.csv")
        if os.path.exists(csv_tmp):
            os.remove(csv_tmp)
        shutil.copy(csv_file_path, csv_tmp)

        def mock_get_file_path(self, name):
            return csv_tmp

        monkeypatch.setattr(MockFileController, "get_file_path", mock_get_file_path)

        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.query(
            Insert(
                table=Identifier(parts=["someTable"]),
                columns=[
                    "col_one",
                    "col_two",
                    "col_three",
                    "col_four",
                ],
                values=[
                    [1, -1, 0.1, "A"],
                    [2, -2, 0.2, "B"],
                    [3, -3, 0.3, "C"],
                ],
            )
        )

        assert response.type == RESPONSE_TYPE.OK

    def test_query_create(self):
        """Test a valid create table query"""
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.query(
            CreateTable(
                name=Identifier(parts=["someTable"]),
                columns=[TableColumn(name="col1"), TableColumn(name="col2")],
            )
        )

        assert response.type == RESPONSE_TYPE.OK

    def test_query_create_or_replace(self):
        """Test a valid create or replace table query"""
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.query(
            CreateTable(
                name=Identifier(parts=["someTable"]),
                columns=[TableColumn(name="col1"), TableColumn(name="col2")],
                is_replace=True,
            )
        )

        assert response.type == RESPONSE_TYPE.OK

    def test_query_bad_type(self):
        """Test an invalid query type for files"""
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.query(Update([Identifier(parts=["someTable"])]))

        assert response.type == RESPONSE_TYPE.ERROR

    def test_native_query(self, monkeypatch):
        """Test a valid native table query"""

        def mock_delete(self, name):
            return True

        monkeypatch.setattr(MockFileController, "delete_file", mock_delete)
        file_handler = FileHandler(file_controller=MockFileController())
        response = file_handler.native_query("DROP TABLE one")

        assert response.type == RESPONSE_TYPE.OK

    def test_invalid_native_query(self):
        file_handler = FileHandler(file_controller=MockFileController())
        with pytest.raises(ParsingException):
            file_handler.native_query("INVALID QUERY")


def test_handle_source():
    def get_reader(file_path):
        # using path
        reader = FileReader(path=file_path)
        yield reader

        # using file descriptor
        with open(file_path, "rb") as fd:
            reader = FileReader(file=fd)
            yield reader
            fd.seek(0)
            content = fd.read()

        # using bytesio
        fd = BytesIO(content)
        reader = FileReader(file=fd, name=Path(file_path).name)
        yield reader

    for file_path, expected_columns in (
        (csv_file(), test_file_content[0]),
        (xlsx_file(), test_file_content[0]),
        (json_file(), test_file_content[0]),
        (parquet_file(), test_file_content[0]),
        (pdf_file(), ["content", "metadata"]),
        (txt_file(), ["content", "metadata"]),
    ):
        # using different methods to create reader
        for reader in get_reader(file_path):
            df = reader.get_page_content()
            assert isinstance(df, pandas.DataFrame)

            assert df.columns.tolist() == expected_columns

            # The pdf and txt files have some different content
            if reader.get_format() not in ("pdf", "txt"):
                assert len(df) == len(test_file_content) - 1
                assert df.values.tolist() == test_file_content[1:]


@pytest.mark.parametrize(
    "csv_string,delimiter",
    [
        (StringIO("example,csv,file"), ","),
        (StringIO("example;csv;file"), ";"),
        (StringIO("example\tcsv\tfile"), "\t"),
    ],
)
def test_check_valid_dialects(csv_string, delimiter):
    dialect = FileReader._get_csv_dialect(csv_string)
    assert dialect.delimiter == delimiter


def test_tsv():
    file = BytesIO(b"example;csv;file\tname")

    reader = FileReader(file=file, name="test.tsv")
    assert reader.get_format() == "csv"
    assert reader.parameters["delimiter"] == "\t"

    df = reader.get_page_content()
    assert len(df.columns) == 2


def test_bad_csv_header():
    file = BytesIO(b" a,b  ,c\n1,2,3\n")
    reader = FileReader(file=file, name="test.tsv")
    df = reader.get_page_content()
    assert set(df.columns) == set(["a", "b", "c"])

    wrong_data = [
        b"a, ,c\n1,2,3\n",
        b"a,  \t,c\n1,2,3\n",
        b"   ,b,c\n1,2,3\n",
    ]
    for data in wrong_data:
        reader = FileReader(file=BytesIO(data), name="test.tsv")
        with pytest.raises(FileProcessingError):
            df = reader.get_page_content()


def test_check_invalid_dialects():
    with pytest.raises(Exception):
        FileHandler._get_csv_dialect("example csv file")
    with pytest.raises(Exception):
        FileHandler._get_csv_dialect("example\ncsv\nfile")
    with pytest.raises(Exception):
        FileHandler._get_csv_dialect("example|csv|file")


def test_get_tables():
    file_handler = FileHandler(file_controller=MockFileController())
    response = file_handler.get_tables()

    assert response.type == RESPONSE_TYPE.TABLE

    expected_df = pandas.DataFrame(
        [{"TABLE_NAME": x[0], "TABLE_ROWS": x[1], "TABLE_TYPE": "BASE TABLE"} for x in file_records]
    )

    assert response.data_frame.equals(expected_df)


def test_get_columns():
    file_handler = FileHandler(file_controller=MockFileController())
    response = file_handler.get_columns("mock")

    assert response.type == RESPONSE_TYPE.TABLE

    expected_df = pandas.DataFrame([{"Field": x, "Type": "str"} for x in file_records[0][2]])

    assert response.data_frame.equals(expected_df)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    url={
        "type": ARG_TYPE.STR,
        "description": "The URI-Like connection string to the MySQL server. If provided, it will override the other connection arguments.",
        "required": False,
        "label": "URL",
    },
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate with the MySQL server.",
        "required": True,
        "label": "User",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password to authenticate the user with the MySQL server.",
        "required": True,
        "label": "Password",
        "secret": True,
    },
    database={
        "type": ARG_TYPE.STR,
        "description": "The database name to use when connecting with the MySQL server.",
        "required": True,
        "label": "Database",
    },
    host={
        "type": ARG_TYPE.STR,
        "description": "The host name or IP address of the MySQL server. NOTE: use '127.0.0.1' instead of 'localhost' to connect to local server.",
        "required": True,
        "label": "Host",
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "The TCP/IP port of the MySQL server. Must be an integer.",
        "required": True,
        "label": "Port",
    },
    ssl={
        "type": ARG_TYPE.BOOL,
        "description": "Set it to True to enable ssl.",
        "required": False,
        "label": "ssl",
    },
    ssl_ca={
        "type": ARG_TYPE.PATH,
        "description": "Path or URL of the Certificate Authority (CA) certificate file",
        "required": False,
        "label": "ssl_ca",
    },
    ssl_cert={
        "type": ARG_TYPE.PATH,
        "description": "Path name or URL of the server public key certificate file",
        "required": False,
        "label": "ssl_cert",
    },
    ssl_key={
        "type": ARG_TYPE.PATH,
        "description": "The path name or URL of the server private key file",
        "required": False,
        "label": "ssl_key",
    },
)

connection_args_example = OrderedDict(
    host="127.0.0.1", port=3306, user="root", password="password", database="database"
)



import os
import base64
import hashlib
import json
import datetime
import textwrap
from functools import wraps
from collections.abc import Callable

from cryptography.fernet import Fernet
from mindsdb_sql_parser.ast import Identifier

from mindsdb.utilities.fs import create_process_mark, delete_process_mark, set_process_mark
from mindsdb.utilities import log
from mindsdb.utilities.config import Config


logger = log.getLogger(__name__)


def get_handler_install_message(handler_name):
    if Config().use_docker_env:
        container_id = os.environ.get("HOSTNAME", "<container_id>")
        return textwrap.dedent(f"""\
            To install the {handler_name} handler, run the following in your terminal outside the docker container
            ({container_id} is the ID of this container):

                docker exec {container_id} pip install 'mindsdb[{handler_name}]'""")
    else:
        return textwrap.dedent(f"""\
            To install the {handler_name} handler, run the following in your terminal:

                pip install 'mindsdb[{handler_name}]'  # If you installed mindsdb via pip
                pip install '.[{handler_name}]'        # If you installed mindsdb from source""")


def cast_row_types(row, field_types):
    """ """
    keys = [x for x in row.keys() if x in field_types]
    for key in keys:
        t = field_types[key]
        if t == "Timestamp" and isinstance(row[key], (int, float)):
            timestamp = datetime.datetime.fromtimestamp(row[key], datetime.timezone.utc)
            row[key] = timestamp.strftime("%Y-%m-%d %H:%M:%S")
        elif t == "Date" and isinstance(row[key], (int, float)):
            timestamp = datetime.datetime.fromtimestamp(row[key], datetime.timezone.utc)
            row[key] = timestamp.strftime("%Y-%m-%d")
        elif t == "Int" and isinstance(row[key], (int, float, str)):
            try:
                logger.debug(f"cast {row[key]} to {int(row[key])}")
                row[key] = int(row[key])
            except Exception:
                pass


def mark_process(name: str, custom_mark: str = None) -> Callable:
    def mark_process_wrapper(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            if custom_mark is None:
                mark = create_process_mark(name)
            else:
                mark = set_process_mark(name, custom_mark)

            try:
                return func(*args, **kwargs)
            finally:
                delete_process_mark(name, mark)

        return wrapper

    return mark_process_wrapper


def init_lexer_parsers():
    from mindsdb_sql_parser.lexer import MindsDBLexer
    from mindsdb_sql_parser.parser import MindsDBParser

    return MindsDBLexer(), MindsDBParser()


def resolve_table_identifier(identifier: Identifier, default_database: str = None) -> tuple:
    parts = identifier.parts

    parts_count = len(parts)
    if parts_count == 1:
        return (None, parts[0])
    elif parts_count == 2:
        return (parts[0], parts[1])
    else:
        raise Exception(f"Table identifier must contain max 2 parts: {parts}")


def resolve_model_identifier(identifier: Identifier) -> tuple:
    """
    Splits a model identifier into its database, model name, and version components.

    The identifier may contain one, two, or three parts.
    The function supports both quoted and unquoted identifiers, and normalizes names to lowercase if unquoted.

    Examples:
        >>> resolve_model_identifier(Identifier(parts=['a', 'b']))
        ('a', 'b', None)
        >>> resolve_model_identifier(Identifier(parts=['a', '1']))
        (None, 'a', 1)
        >>> resolve_model_identifier(Identifier(parts=['a']))
        (None, 'a', None)
        >>> resolve_model_identifier(Identifier(parts=['a', 'b', 'c']))
        (None, None, None)  # not found

    Args:
        identifier (Identifier): The identifier object containing parts and is_quoted attributes.

    Returns:
        tuple: (database_name, model_name, model_version)
            - database_name (str or None): The name of the database/project, or None if not specified.
            - model_name (str or None): The name of the model, or None if not found.
            - model_version (int or None): The model version as an integer, or None if not specified.
    """
    model_name = None
    db_name = None
    version = None
    model_name_quoted = None
    db_name_quoted = None

    match identifier.parts, identifier.is_quoted:
        case [model_name], [model_name_quoted]:
            ...
        case [model_name, str(version)], [model_name_quoted, _] if version.isdigit():
            ...
        case [model_name, int(version)], [model_name_quoted, _]:
            ...
        case [db_name, model_name], [db_name_quoted, model_name_quoted]:
            ...
        case [db_name, model_name, str(version)], [db_name_quoted, model_name_quoted, _] if version.isdigit():
            ...
        case [db_name, model_name, int(version)], [db_name_quoted, model_name_quoted, _]:
            ...
        case [db_name, model_name, str(version)], [db_name_quoted, model_name_quoted, _]:
            # for back compatibility. May be delete?
            return (None, None, None)
        case _:
            ...  # may be raise ValueError?

    if model_name_quoted is False:
        model_name = model_name.lower()

    if db_name_quoted is False:
        db_name = db_name.lower()

    if isinstance(version, int) or isinstance(version, str) and version.isdigit():
        version = int(version)
    else:
        version = None

    return db_name, model_name, version


def encrypt(string: bytes, key: str) -> bytes:
    hashed_string = hashlib.sha256(key.encode()).digest()

    fernet_key = base64.urlsafe_b64encode(hashed_string)

    cipher = Fernet(fernet_key)
    return cipher.encrypt(string)


def decrypt(encripted: bytes, key: str) -> bytes:
    hashed_string = hashlib.sha256(key.encode()).digest()

    fernet_key = base64.urlsafe_b64encode(hashed_string)

    cipher = Fernet(fernet_key)
    return cipher.decrypt(encripted)


def encrypt_json(data: dict, key: str) -> bytes:
    json_str = json.dumps(data)
    return encrypt(json_str.encode(), key)


def decrypt_json(encrypted_data: bytes, key: str) -> dict:
    decrypted = decrypt(encrypted_data, key)
    return json.loads(decrypted)



import os
import copy

from duckdb.typing import BIGINT, DOUBLE, VARCHAR, BLOB, BOOLEAN
from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.utilities.config import config


def python_to_duckdb_type(py_type):
    if py_type == "int":
        return BIGINT
    elif py_type == "float":
        return DOUBLE
    elif py_type == "str":
        return VARCHAR
    elif py_type == "bool":
        return BOOLEAN
    elif py_type == "bytes":
        return BLOB
    else:
        # Unknown
        return VARCHAR


# duckdb doesn't like *args
def function_maker(n_args, other_function):
    return [
        lambda: other_function(),
        lambda arg_0: other_function(arg_0),
        lambda arg_0, arg_1: other_function(arg_0, arg_1),
        lambda arg_0, arg_1, arg_2: other_function(arg_0, arg_1, arg_2),
        lambda arg_0, arg_1, arg_2, arg_3: other_function(arg_0, arg_1, arg_2, arg_2),
    ][n_args]


class BYOMFunctionsController:
    """
    User functions based on BYOM handler
    """

    def __init__(self, session):
        self.session = session

        self.byom_engines = None
        self.byom_methods = {}
        self.byom_handlers = {}

        self.callbacks = {}

    def get_engines(self):
        # get all byom engines
        if self.byom_engines is None:
            # first run
            self.byom_engines = []
            for name, info in self.session.integration_controller.get_all().items():
                if info["type"] == "ml" and info["engine"] == "byom":
                    if info["connection_data"].get("mode") == "custom_function":
                        self.byom_engines.append(name)
        return self.byom_engines

    def get_methods(self, engine):
        if engine not in self.byom_methods:
            ml_handler = self.session.integration_controller.get_ml_handler(engine)

            storage = HandlerStorage(ml_handler.integration_id)
            methods = storage.json_get("methods")
            self.byom_methods[engine] = methods
            self.byom_handlers[engine] = ml_handler

        return self.byom_methods[engine]

    def check_function(self, node):
        engine = node.namespace
        if engine not in self.get_engines():
            return

        methods = self.get_methods(engine)

        fnc_name = node.op.lower()
        if fnc_name not in methods:
            # do nothing
            return

        new_name = f"{node.namespace}_{fnc_name}"
        node.op = new_name

        if new_name in self.callbacks:
            # already exists
            return self.callbacks[new_name]

        def callback(*args):
            return self.method_call(engine, fnc_name, args)

        input_types = [param["type"] for param in methods[fnc_name]["input_params"]]

        meta = {
            "name": new_name,
            "callback": callback,
            "input_types": input_types,
            "output_type": methods[fnc_name]["output_type"],
        }

        self.callbacks[new_name] = meta
        return meta

    def method_call(self, engine, method_name, args):
        return self.byom_handlers[engine].function_call(method_name, args)

    def create_function_set(self):
        return DuckDBFunctions(self)


class FunctionController(BYOMFunctionsController):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def check_function(self, node):
        meta = super().check_function(node)
        if meta is not None:
            return meta

        # builtin functions
        if node.op.lower() == "llm":
            return self.llm_call_function(node)

        elif node.op.lower() == "to_markdown":
            return self.to_markdown_call_function(node)

    def llm_call_function(self, node):
        name = node.op.lower()

        if name in self.callbacks:
            return self.callbacks[name]

        chat_model_params = self._parse_chat_model_params()

        try:
            from langchain_core.messages import HumanMessage
            from mindsdb.interfaces.agents.langchain_agent import create_chat_model

            llm = create_chat_model(chat_model_params)
        except Exception as e:
            raise RuntimeError(f"Unable to use LLM function, check ENV variables: {e}")

        def callback(question):
            resp = llm([HumanMessage(question)])
            return resp.content

        meta = {"name": name, "callback": callback, "input_types": ["str"], "output_type": "str"}
        self.callbacks[name] = meta
        return meta

    def to_markdown_call_function(self, node):
        # load on-demand because lib is heavy
        from mindsdb.interfaces.functions.to_markdown import ToMarkdown

        name = node.op.lower()

        if name in self.callbacks:
            return self.callbacks[name]

        def prepare_chat_model_params(chat_model_params: dict) -> dict:
            """
            Parepares the chat model parameters for the ToMarkdown function.
            """
            params_copy = copy.deepcopy(chat_model_params)
            params_copy["model"] = params_copy.pop("model_name")

            # Set the base_url for the Google provider.
            if params_copy["provider"] == "google" and "base_url" not in params_copy:
                params_copy["base_url"] = "https://generativelanguage.googleapis.com/v1beta/"

            params_copy.pop("api_keys")
            params_copy.pop("provider")

            return params_copy

        def callback(file_path_or_url):
            chat_model_params = self._parse_chat_model_params("TO_MARKDOWN_FUNCTION_")
            chat_model_params = prepare_chat_model_params(chat_model_params)

            to_markdown = ToMarkdown()
            return to_markdown.call(file_path_or_url, **chat_model_params)

        meta = {"name": name, "callback": callback, "input_types": ["str"], "output_type": "str"}
        self.callbacks[name] = meta
        return meta

    def _parse_chat_model_params(self, param_prefix: str = "LLM_FUNCTION_"):
        """
        Parses the environment variables for chat model parameters.
        """
        chat_model_params = config.get("default_llm") or {}
        for k, v in os.environ.items():
            if k.startswith(param_prefix):
                param_name = k[len(param_prefix) :]
                if param_name == "MODEL":
                    chat_model_params["model_name"] = v
                else:
                    chat_model_params[param_name.lower()] = v

        if "provider" not in chat_model_params:
            chat_model_params["provider"] = "openai"

        if "api_key" in chat_model_params:
            # move to api_keys dict
            chat_model_params["api_keys"] = {chat_model_params["provider"]: chat_model_params["api_key"]}

        return chat_model_params


class DuckDBFunctions:
    def __init__(self, controller):
        self.controller = controller
        self.functions = {}

    def check_function(self, node):
        meta = self.controller.check_function(node)
        if meta is None:
            return

        name = meta["name"]

        if name in self.functions:
            return

        input_types = [python_to_duckdb_type(param) for param in meta["input_types"]]

        self.functions[name] = {
            "callback": function_maker(len(input_types), meta["callback"]),
            "input": input_types,
            "output": python_to_duckdb_type(meta["output_type"]),
        }

    def register(self, connection):
        for name, info in self.functions.items():
            connection.create_function(name, info["callback"], info["input"], info["output"], null_handling="special")



import asyncio
from collections import namedtuple
from typing import Any, Dict, List, Optional

from mindsdb.interfaces.agents.langchain_agent import create_chat_model
from langchain.chains.base import Chain
from langchain.chains.combine_documents.stuff import StuffDocumentsChain
from langchain.chains.llm import LLMChain
from langchain.chains.combine_documents.map_reduce import MapReduceDocumentsChain, ReduceDocumentsChain
from langchain_core.callbacks import dispatch_custom_event
from langchain_core.callbacks.manager import CallbackManagerForChainRun
from langchain_core.documents import Document
from langchain_core.prompts import PromptTemplate
from pandas import DataFrame

from mindsdb.integrations.libs.vectordatabase_handler import VectorStoreHandler
from mindsdb.integrations.utilities.rag.settings import SummarizationConfig
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator
from mindsdb.utilities import log

logger = log.getLogger(__name__)

Summary = namedtuple('Summary', ['source_id', 'content'])


def create_map_reduce_documents_chain(summarization_config: SummarizationConfig, input: str) -> ReduceDocumentsChain:
    """Creates a chain that map-reduces documents into a single consolidated summary."""
    summarization_llm = create_chat_model({
        'model_name': summarization_config.llm_config.model_name,
        'provider': summarization_config.llm_config.provider,
        **summarization_config.llm_config.params
    })

    reduce_prompt_template = summarization_config.reduce_prompt_template
    reduce_prompt = PromptTemplate.from_template(reduce_prompt_template)
    if 'input' in reduce_prompt.input_variables:
        reduce_prompt = reduce_prompt.partial(input=input)

    reduce_chain = LLMChain(llm=summarization_llm, prompt=reduce_prompt)

    combine_documents_chain = StuffDocumentsChain(
        llm_chain=reduce_chain,
        document_variable_name='docs'
    )

    return ReduceDocumentsChain(
        combine_documents_chain=combine_documents_chain,
        collapse_documents_chain=combine_documents_chain,
        token_max=summarization_config.max_summarization_tokens
    )


class LocalContextSummarizerChain(Chain):
    """Summarizes M chunks before and after a given chunk in a document."""

    doc_id_key: str = 'original_row_id'
    chunk_index_key: str = 'chunk_index'

    vector_store_handler: VectorStoreHandler
    table_name: str = 'embeddings'
    content_column_name: str = 'content'
    metadata_column_name: str = 'metadata'

    summarization_config: SummarizationConfig
    map_reduce_documents_chain: Optional[ReduceDocumentsChain] = None

    def _select_chunks_from_vector_store(self, doc_id: str) -> DataFrame:
        condition = FilterCondition(
            f"{self.metadata_column_name}->>'{self.doc_id_key}'",
            FilterOperator.EQUAL,
            doc_id
        )
        return self.vector_store_handler.select(
            self.table_name,
            columns=[self.content_column_name, self.metadata_column_name],
            conditions=[condition]
        )

    async def _get_all_chunks_for_document(self, doc_id: str) -> List[Document]:
        df = await asyncio.get_event_loop().run_in_executor(
            None, self._select_chunks_from_vector_store, doc_id
        )
        chunks = []
        for _, row in df.iterrows():
            metadata = row.get(self.metadata_column_name, {})
            metadata[self.chunk_index_key] = row.get('chunk_id', 0)
            chunks.append(Document(page_content=row[self.content_column_name], metadata=metadata))

        return sorted(chunks, key=lambda x: x.metadata.get(self.chunk_index_key, 0))

    async def summarize_local_context(self, doc_id: str, target_chunk_index: int, M: int) -> Summary:
        """
        Summarizes M chunks before and after the given chunk.

        Args:
            doc_id (str): Document ID.
            target_chunk_index (int): Index of the chunk to summarize around.
            M (int): Number of chunks before and after to include.

        Returns:
            Summary: Summary object containing source_id and summary content.
        """
        logger.debug(f"Fetching chunks for document {doc_id}")
        all_chunks = await self._get_all_chunks_for_document(doc_id)

        if not all_chunks:
            logger.warning(f"No chunks found for document {doc_id}")
            return Summary(source_id=doc_id, content='')

        # Determine window boundaries
        start_idx = max(0, target_chunk_index - M)
        end_idx = min(len(all_chunks), target_chunk_index + M + 1)
        local_chunks = all_chunks[start_idx:end_idx]

        logger.debug(f"Summarizing chunks {start_idx} to {end_idx - 1} for document {doc_id}")

        if not self.map_reduce_documents_chain:
            self.map_reduce_documents_chain = create_map_reduce_documents_chain(
                self.summarization_config, input="Summarize these chunks."
            )

        summary_result = await self.map_reduce_documents_chain.ainvoke(local_chunks)
        summary_text = summary_result.get('output_text', '')

        logger.debug(f"Generated summary: {summary_text[:100]}...")

        return Summary(source_id=doc_id, content=summary_text)

    @property
    def input_keys(self) -> List[str]:
        return [self.context_key, self.question_key]

    @property
    def output_keys(self) -> List[str]:
        return [self.context_key, self.question_key]

    async def _get_source_summary(self, source_id: str, map_reduce_documents_chain: MapReduceDocumentsChain) -> Summary:
        if not source_id:
            logger.warning("Received empty source_id, returning empty summary")
            return Summary(source_id='', content='')

        logger.debug(f"Getting summary for source ID: {source_id}")
        source_chunks = await self._get_all_chunks_for_document(source_id)

        if not source_chunks:
            logger.warning(f"No chunks found for source ID: {source_id}")
            return Summary(source_id=source_id, content='')

        logger.debug(f"Summarizing {len(source_chunks)} chunks for source ID: {source_id}")
        summary = await map_reduce_documents_chain.ainvoke(source_chunks)
        content = summary.get('output_text', '')
        logger.debug(f"Generated summary for source ID {source_id}: {content[:100]}...")

        # Stream summarization update.
        dispatch_custom_event('summary', {'source_id': source_id, 'content': content})

        return Summary(source_id=source_id, content=content)

    async def _get_source_summaries(self, source_ids: List[str], map_reduce_documents_chain: MapReduceDocumentsChain) -> \
            List[Summary]:
        summaries = await asyncio.gather(
            *[self._get_source_summary(source_id, map_reduce_documents_chain) for source_id in source_ids]
        )
        return summaries

    def _call(
            self,
            inputs: Dict[str, Any],
            run_manager: Optional[CallbackManagerForChainRun] = None
    ) -> Dict[str, Any]:
        # Step 1: Connect to vector store to ensure embeddings are accessible
        self.vector_store_handler.connect()

        context_chunks: List[Document] = inputs.get(self.context_key, [])
        logger.debug(f"Found {len(context_chunks)} context chunks.")

        # Step 2: Extract unique document IDs from the provided chunks
        unique_document_ids = self._get_document_ids_from_chunks(context_chunks)
        logger.debug(f"Extracted {len(unique_document_ids)} unique document IDs: {unique_document_ids}")

        # Step 3: Initialize the summarization chain if not provided
        question = inputs.get(self.question_key, '')
        map_reduce_documents_chain = self.map_reduce_documents_chain or create_map_reduce_documents_chain(
            self.summarization_config, question
        )

        # Step 4: Dispatch event to signal summarization start
        if run_manager:
            run_manager.on_text("Starting summarization for documents.", verbose=True)

        # Step 5: Process each document ID to summarize chunks with local context
        for doc_id in unique_document_ids:
            logger.debug(f"Fetching and summarizing chunks for document ID: {doc_id}")

            # Fetch all chunks for the document
            chunks = asyncio.get_event_loop().run_until_complete(self._get_all_chunks_for_document(doc_id))
            if not chunks:
                logger.warning(f"No chunks found for document ID: {doc_id}")
                continue

            # Summarize each chunk with M neighboring chunks
            M = self.neighbor_window
            for i, chunk in enumerate(chunks):
                window_chunks = chunks[max(0, i - M): min(len(chunks), i + M + 1)]
                local_summary = asyncio.get_event_loop().run_until_complete(
                    map_reduce_documents_chain.ainvoke(window_chunks)
                )
                chunk.metadata['summary'] = local_summary.get('output_text', '')
                logger.debug(f"Chunk {i} summary: {chunk.metadata['summary'][:100]}...")

        # Step 6: Update the original context chunks with the newly generated summaries
        for chunk in context_chunks:
            doc_id = str(chunk.metadata.get(self.doc_id_key, ''))
            matching_chunk = next((c for c in chunks if c.metadata.get(self.doc_id_key) == doc_id and c.metadata.get(
                'chunk_index') == chunk.metadata.get('chunk_index')), None)
            if matching_chunk:
                chunk.metadata['summary'] = matching_chunk.metadata.get('summary', '')
            else:
                chunk.metadata['summary'] = ''
                logger.warning(f"No matching chunk found for doc_id: {doc_id}")

        # Step 7: Signal summarization end
        if run_manager:
            run_manager.on_text("Summarization completed.", verbose=True)

        logger.debug(f"Updated {len(context_chunks)} context chunks with summaries.")
        return inputs



from typing import List, Optional, Union

from mindsdb.integrations.libs.api_handler import MetaAPIHandler
from mindsdb.integrations.libs.base import MetaDatabaseHandler
from mindsdb.utilities import log


logger = log.getLogger("mindsdb")


class BaseDataCatalog:
    """
    This is the base class for the Data Catalog interface.
    """

    def __init__(self, database_name: str, table_names: Optional[List[str]] = None) -> None:
        """
        Initialize the DataCatalogReader.

        Args:
            database_name (str): The data source to read/write metadata from.
            table_names (Optional[List[str]]): The list of table names to read or write metadata for. If None, all tables will be read or written.
        """
        from mindsdb.api.executor.controllers.session_controller import (
            SessionController,
        )

        session = SessionController()

        self.database_name = database_name
        self.data_handler: Union[MetaDatabaseHandler, MetaAPIHandler] = session.integration_controller.get_data_handler(
            database_name
        )
        integration = session.integration_controller.get(database_name)
        self.integration_id = integration["id"]
        self.integration_engine = integration["engine"]
        # TODO: Handle situations where a schema is provided along with the database name, e.g., 'schema.table'.
        # TODO: Handle situations where a file path is provided with integrations like S3, e.g., 'dir/file.csv'.
        self.table_names = table_names

        self.logger = logger

    def is_data_catalog_supported(self) -> bool:
        """
        Check if the data catalog is supported for the given database.

        Returns:
            bool: True if the data catalog is supported, False otherwise.
        """
        if not isinstance(self.data_handler, (MetaDatabaseHandler, MetaAPIHandler)):
            self.logger.warning(f"Data catalog is not supported for the '{self.integration_engine}' integration'. ")
            return False

        return True



import unittest
from unittest.mock import patch

from mindsdb_sql_parser import ast
from mindsdb_sql_parser.ast import BinaryOperation, Identifier, Constant
from mindsdb_sql_parser.ast.select.star import Star
from mindsdb.integrations.handlers.discord_handler.discord_handler import DiscordHandler


class DiscordHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.handler = DiscordHandler(
            name='discord_datasource', connection_data={'token': 'test-discord-token'}
        )

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    @patch('mindsdb.integrations.handlers.discord_handler.discord_handler.requests.get')
    def test_1_read_messages(self, mock_get):
        mock_get.return_value.status_code = 200
        mock_get.return_value.json.return_value = [{'content': 'Test message'}]
        query = ast.Select(
            targets=[Star()],
            from_table="messages",
            where=BinaryOperation(
                op='=', args=[Identifier('channel_id'), Constant('1234567890')]
            ),
        )

        self.handler._tables['messages'].select(query)
        mock_get.assert_called_with(
            'https://discord.com/api/v10/channels/1234567890/messages',
            headers={
                'Authorization': 'Bot test-discord-token',
                'Content-Type': 'application/json',
            },
            params={'limit': 100},
        )

    @patch(
        'mindsdb.integrations.handlers.discord_handler.discord_handler.requests.post'
    )
    def test_2_send_message(self, mock_post):
        mock_post.return_value.status_code = 200
        self.handler._tables['messages'].send_message(
            [{'channel_id': '1234567890', 'text': 'Test message'}]
        )
        mock_post.assert_called_with(
            'https://discord.com/api/v10/channels/1234567890/messages',
            headers={
                'Authorization': 'Bot test-discord-token',
                'Content-Type': 'application/json',
            },
            json={'content': 'Test message'},
        )


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import asyncio
import logging
from typing import Any, Dict, Optional, Sequence

from langchain.retrievers.document_compressors.base import BaseDocumentCompressor
from langchain_core.callbacks import Callbacks, dispatch_custom_event
from langchain_core.documents import Document

from mindsdb.integrations.utilities.rag.rerankers.base_reranker import BaseLLMReranker

log = logging.getLogger(__name__)


class LLMReranker(BaseDocumentCompressor, BaseLLMReranker):
    remove_irrelevant: bool = True  # New flag to control removal of irrelevant documents

    def _dispatch_rerank_event(self, data):
        dispatch_custom_event("rerank", data)

    async def acompress_documents(
        self,
        documents: Sequence[Document],
        query: str,
        callbacks: Optional[Callbacks] = None,
    ) -> Sequence[Document]:
        """Async compress documents using reranking with proper error handling."""
        if callbacks:
            await callbacks.on_retriever_start({"query": query}, "Reranking documents")

        log.info(f"Async compressing documents. Initial count: {len(documents)}")
        if not documents:
            if callbacks:
                await callbacks.on_retriever_end({"documents": []})
            return []

        # Stream reranking update.
        dispatch_custom_event('rerank_begin', {'num_documents': len(documents)})

        try:
            # Prepare query-document pairs
            query_document_pairs = [(query, doc.page_content) for doc in documents]

            if callbacks:
                await callbacks.on_text("Starting document reranking...")

            # Get ranked results
            ranked_results = await self._rank(query_document_pairs, rerank_callback=self._dispatch_rerank_event)

            # Sort by score in descending order
            ranked_results.sort(key=lambda x: x[1], reverse=True)

            # Filter based on threshold and num_docs_to_keep
            filtered_docs = []
            for doc, score in ranked_results:
                if score >= self.filtering_threshold:
                    matching_doc = next(d for d in documents if d.page_content == doc)
                    matching_doc.metadata = {**(matching_doc.metadata or {}), "relevance_score": score}
                    filtered_docs.append(matching_doc)

                    if callbacks:
                        await callbacks.on_text(f"Document scored {score:.2f}")

                    if self.num_docs_to_keep and len(filtered_docs) >= self.num_docs_to_keep:
                        break

            log.info(f"Async compression complete. Final count: {len(filtered_docs)}")

            if callbacks:
                await callbacks.on_retriever_end({"documents": filtered_docs})

            return filtered_docs

        except Exception as e:
            error_msg = f"Error during async document compression: {str(e)}"
            log.error(error_msg)
            if callbacks:
                await callbacks.on_retriever_error(error_msg)
            return documents  # Return original documents on error

    def compress_documents(
        self,
        documents: Sequence[Document],
        query: str,
        callbacks: Optional[Callbacks] = None,
    ) -> Sequence[Document]:
        """Sync wrapper for async compression."""
        return asyncio.run(self.acompress_documents(documents, query, callbacks))

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Get the identifying parameters."""
        return {
            "model": self.model,
            "temperature": self.temperature,
            "remove_irrelevant": self.remove_irrelevant,
            "method": self.method,
        }



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Ingres server.'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password to authenticate the user with the Ingres server.',
        'secret': True
    },
    server={
        'type': ARG_TYPE.STR,
        'description': 'The server used to authenticate with the Ingres server.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specify database name to connect Ingres server'
    },
    servertype={
        'type': ARG_TYPE.STR,
        'description': 'Specify server type to connect Ingres server'
    }
)

connection_args_example = OrderedDict(
    user='admin',
    password='password',
    server='(local)',
    database='test_db',
    servertype='ingres'
)



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The hostname, IP address, or URL of the Teradata server.',
        'required': True,
        'label': 'Host'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The username for the Teradata database.',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'The password for the Teradata database.',
        'secret': True,
        'required': True,
        'label': 'Password'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': "The name of the Teradata database to connect to. Defaults is the user's default database.",
        'required': False,
        'label': 'Database'
    }
)

connection_args_example = OrderedDict(
    host='192.168.0.41',
    user='dbc',
    password='dbc',
    database='HR'
)



from typing import Dict, List, Text

from mindsdb_sql_parser.ast import Select, Star, Identifier
import pandas as pd
from salesforce_api.exceptions import RestRequestCouldNotBeUnderstoodError

from mindsdb.integrations.libs.api_handler import MetaAPIResource
from mindsdb.integrations.utilities.sql_utils import FilterCondition, FilterOperator
from mindsdb.utilities import log


logger = log.getLogger(__name__)


def create_table_class(resource_name: Text) -> MetaAPIResource:
    """
    Creates a table class for the given Salesforce resource.
    """

    class AnyTable(MetaAPIResource):
        """
        This is the table abstraction for any resource of the Salesforce API.
        """

        def __init__(self, *args, table_name=None, **kwargs):
            """
            Initializes the AnyTable class.

            Args:
                *args: Variable length argument list.
                table_name (str): The name of the table that represents the Salesforce resource.
                **kwargs: Arbitrary keyword arguments.
            """
            super().__init__(*args, table_name=table_name, **kwargs)
            self.resource_metadata = None

        def select(self, query: Select) -> pd.DataFrame:
            """
            Executes a SELECT SQL query represented by an ASTNode object on the Salesforce resource and retrieves the data (if any).

            Args:
                query (ASTNode): An ASTNode object representing the SQL query to be executed.

            Returns:
                pd.DataFrame: A DataFrame containing the data retrieved from the Salesforce resource.
            """
            query.from_table = resource_name

            # SOQL does not support * in SELECT queries. Replace * with column names.
            if isinstance(query.targets[0], Star):
                query.targets = [Identifier(column) for column in self.get_columns()]

            # SOQL does not support column aliases. Remove column aliases.
            column_aliases = {}
            for column in query.targets:
                if column.alias is not None:
                    column_aliases[column.parts[-1]] = column.alias.parts[-1]
                    column.alias = None

            client = self.handler.connect()

            query_str = query.to_string()

            # SOQL does not support backticks. Remove backticks.
            query_str = query_str.replace("`", "")
            results = client.sobjects.query(query_str)

            for result in results:
                del result["attributes"]

            df = pd.DataFrame(results)
            df.rename(columns=column_aliases, inplace=True)

            return df

        def add(self, item: Dict) -> None:
            """
            Adds a new item to the Salesforce resource.

            Args:
                contact (Dict): The data to be inserted into the Salesforce resource.
            """
            client = self.handler.connect()
            getattr(client.sobjects, resource_name).insert(item)

        def modify(self, conditions: List[FilterCondition], values: Dict) -> None:
            """
            Modifies items in the Salesforce resource based on the specified conditions.

            Args:
                conditions (List[FilterCondition]): The conditions based on which the items are to be modified.
                values (Dict): The values to be updated in the items.
            """
            client = self.handler.connect()

            ids = self._validate_conditions(conditions)

            for id in ids:
                getattr(client.sobjects, resource_name).update(id, values)

        def remove(self, conditions: List[FilterCondition]) -> None:
            """
            Removes items from the Salesforce resource based on the specified conditions.

            Args:
                conditions (List[FilterCondition]): The conditions based on which the items are to be removed.
            """
            client = self.handler.connect()

            ids = self._validate_conditions(conditions)

            for id in ids:
                getattr(client.sobjects, resource_name).delete(id)

        def _validate_conditions(self, conditions: List[FilterCondition]) -> None:
            """
            Validates the conditions used for filtering items in the Salesforce resource.

            Args:
                conditions (List[FilterCondition]): The conditions to be validated.
            """
            # Salesforce API does not support filtering items based on attributes other than 'Id'. Raise an error if any other column is used.
            if len(conditions) != 1 or conditions[0].column != "Id":
                raise ValueError("Only the 'Id' column can be used to filter items.")

            # Only the 'equals' and 'in' operators can be used on the 'Id' column for deletion. Raise an error if any other operator is used.
            if conditions[0].op not in [FilterOperator.EQUAL, FilterOperator.IN]:
                raise ValueError("Only the 'equals' and 'in' operators can be used on the 'Id' column.")

            return conditions[0].value if isinstance(conditions[0].value, list) else [conditions[0].value]

        def _get_resource_metadata(self) -> Dict:
            """
            Retrieves metadata about the Salesforce resource.

            Returns:
                Dict: A dictionary containing metadata about the Salesforce resource.
            """
            if self.resource_metadata:
                return self.resource_metadata

            client = self.handler.connect()
            return getattr(client.sobjects, resource_name).describe()

        def get_columns(self) -> List[Text]:
            """
            Retrieves the attributes (columns) of the Salesforce resource.

            Returns:
                List[Text]: A list of Attributes (columns) of the Salesforce resource.
            """
            return [field["name"] for field in self._get_resource_metadata()["fields"]]

        def meta_get_tables(self, table_name: str, main_metadata: Dict) -> Dict:
            """
            Retrieves table metadata for the Salesforce resource.

            Args:
                table_name (str): The name given to the table that represents the Salesforce resource.
                main_metadata (Dict): The main metadata dictionary containing information about all Salesforce resources.

            Returns:
                Dict: A dictionary containing table metadata for the Salesforce resource.
            """
            client = self.handler.connect()

            try:
                resource_metadata = next(
                    (resource for resource in main_metadata if resource["name"].lower() == resource_name),
                )
            except Exception as e:
                logger.warning(f"Failed to get resource metadata for {resource_name}: {e}")
                return {
                    "table_name": table_name,
                    "table_type": "BASE TABLE",
                    "table_description": "",
                    "row_count": None,
                }
            # Get row count if Id column is aggregatable.
            row_count = None
            # if next(field for field in resource_metadata['fields'] if field['name'] == 'Id').get('aggregatable', False):
            try:
                row_count = client.sobjects.query(f"SELECT COUNT(Id) FROM {resource_name}")[0]["expr0"]
            except RestRequestCouldNotBeUnderstoodError as request_error:
                logger.warning(f"Failed to get row count for {resource_name}: {request_error}")

            return {
                "table_name": table_name,
                "table_type": "BASE TABLE",
                "table_description": resource_metadata.get("label", ""),
                "row_count": row_count,
            }

        def meta_get_columns(self, table_name: str) -> List[Dict]:
            """
            Retrieves column metadata for the Salesforce resource.

            Args:
                table_name (str): The name given to the table that represents the Salesforce resource.

            Returns:
                List[Dict]: A list of dictionaries containing column metadata for the Salesforce resource.
            """
            resource_metadata = self._get_resource_metadata()

            column_metadata = []
            for field in resource_metadata["fields"]:
                column_metadata.append(
                    {
                        "table_name": table_name,
                        "column_name": field["name"],
                        "data_type": field["type"],
                        "is_nullable": field.get("nillable", False),
                        "default_value": field.get("defaultValue", ""),
                        "description": field.get("inlineHelpText", ""),
                    }
                )

            return column_metadata

        def meta_get_primary_keys(self, table_name: str) -> List[Dict]:
            """
            Retrieves the primary keys for the Salesforce resource.

            Args:
                table_name (str): The name given to the table that represents the Salesforce resource.

            Returns:
                List[Dict]: A list of dictionaries containing primary key metadata for the Salesforce resource.
            """
            return [
                {
                    "table_name": table_name,
                    "column_name": "Id",
                }
            ]

        def meta_get_foreign_keys(self, table_name: str, all_tables: List[str]) -> List[Dict]:
            """
            Retrieves the foreign keys for the Salesforce resource.

            Args:
                table_name (str): The name given to the table that represents the Salesforce resource.
                all_tables (List[str]): A list of all table names in the Salesforce database.

            Returns:
                List[Dict]: A list of dictionaries containing foreign key metadata for the Salesforce resource.
            """
            resource_metadata = self._get_resource_metadata()

            foreign_key_metadata = []
            for child_relationship in resource_metadata.get("childRelationships", []):
                # Skip if the child relationship is not one of the supported tables.
                child_table_name = child_relationship["childSObject"]
                if child_table_name not in all_tables:
                    continue

                foreign_key_metadata.append(
                    {
                        "parent_table_name": table_name,
                        "parent_column_name": "Id",
                        "child_table_name": child_table_name,
                        "child_column_name": child_relationship["field"],
                    }
                )

            return foreign_key_metadata

    return AnyTable



import requests
import textwrap

from mindsdb.utilities.config import Config


def get_aws_meta_data() -> dict:
    ''' returns aws metadata for current instance

        Returns:
            dict: aws metadata
    '''
    aws_meta_data = {
        'public-hostname': None,
        'ami-id': None,
        'instance-id': None
    }
    aws_token = requests.put("http://169.254.169.254/latest/api/token", headers={'X-aws-ec2-metadata-token-ttl-seconds': '30'}).text
    for key in aws_meta_data.keys():
        resp = requests.get(
            f'http://169.254.169.254/latest/meta-data/{key}',
            headers={'X-aws-ec2-metadata-token': aws_token},
            timeout=1
        )
        if resp.status_code != 200:
            continue
        aws_meta_data[key] = resp.text
    if aws_meta_data['instance-id'] is None:
        raise Exception('That is not an AWS environment')
    return aws_meta_data


def register_oauth_client():
    ''' register new oauth client if it is not existed
    '''
    config = Config()
    aws_meta_data = get_aws_meta_data()

    current_aws_meta_data = config.get('aws_meta_data', {})
    oauth_meta = config.get('auth', {}).get('oauth')
    if oauth_meta is None:
        return

    public_hostname = aws_meta_data['public-hostname']
    if (
        current_aws_meta_data.get('public-hostname') != public_hostname
        or oauth_meta.get('client_id') is None
    ):
        resp = requests.post(
            f'https://{oauth_meta["server_host"]}/auth/register_client',
            json={
                'client_name': f'aws_marketplace_{public_hostname}',
                'client_uri': public_hostname,
                'grant_types': 'authorization_code',
                'redirect_uris': textwrap.dedent(f'''
                    https://{public_hostname}/api/auth/callback
                    https://{public_hostname}/api/auth/callback/cloud_home
                '''),
                'response_types': 'code',
                'scope': 'openid profile aws_marketplace',
                'token_endpoint_auth_method': 'client_secret_basic'
            },
            timeout=10
        )

        if resp.status_code != 200:
            raise Exception(f'Wrong answer from auth server: {resp.status_code}, {resp.text}')
        keys = resp.json()
        Config().update({
            'aws_meta_data': aws_meta_data,
            'auth': {
                'oauth': {
                    'client_id': keys['client_id'],
                    'client_secret': keys['client_secret']
                }
            }
        })



import os
from typing import Dict

from mindsdb.interfaces.storage.model_fs import HandlerStorage
from mindsdb.utilities.config import Config

"""Contains utilities to be used by handlers."""


def get_api_key(
    api_name: str,
    create_args: Dict[str, str],
    engine_storage: HandlerStorage = None,
    strict: bool = True,
):
    """Gets the API key needed to use an ML Handler.

    Args:
        api_name (str): Name of the API (e.g. openai, anthropic)
        create_args (Dict[str, str]): Args user passed to the created model with USING keyword
        engine_storage (HandlerStorage): Engine storage for the ML handler
        strict (bool): Whether or not to require the API key

    Returns:
        api_key (str): The API key

    API_KEY preference order:
        1. provided at inference
        2. provided at model creation
        3. provided at engine creation
        4. api key env variable
        5. api_key setting in config.json
    """
    # Special case for vLLM - always return dummy key
    if api_name == "vllm":
        return "EMPTY"

    # 1
    if "using" in create_args and f"{api_name.lower()}_api_key" in create_args["using"]:
        api_key = create_args["using"][f"{api_name.lower()}_api_key"]
        if api_key:
            return api_key

    # 1.5 - Check for generic api_key in using
    if "using" in create_args and "api_key" in create_args["using"]:
        api_key = create_args["using"]["api_key"]
        if api_key:
            return api_key

    # 2
    if f"{api_name.lower()}_api_key" in create_args:
        api_key = create_args[f"{api_name.lower()}_api_key"]
        if api_key:
            return api_key

    # 2.5 - Check for generic api_key
    if "api_key" in create_args:
        api_key = create_args["api_key"]
        if api_key:
            return api_key

    # 3 - Check in params dictionary if it exists (for agents)
    if "params" in create_args and create_args["params"] is not None:
        if f"{api_name.lower()}_api_key" in create_args["params"]:
            api_key = create_args["params"][f"{api_name.lower()}_api_key"]
            if api_key:
                return api_key
        # 3.5 - Check for generic api_key in params
        if "api_key" in create_args["params"]:
            api_key = create_args["params"]["api_key"]
            if api_key:
                return api_key

    # 4
    if engine_storage is not None:
        connection_args = engine_storage.get_connection_args()
        if f"{api_name.lower()}_api_key" in connection_args:
            api_key = connection_args[f"{api_name.lower()}_api_key"]
            if api_key:
                return api_key
        # 4.5 - Check for generic api_key in connection_args
        if "api_key" in connection_args:
            api_key = connection_args["api_key"]
            if api_key:
                return api_key

    # 5
    api_key = os.getenv(f"{api_name.lower()}_api_key")
    if api_key:
        return api_key
    api_key = os.getenv(f"{api_name.upper()}_API_KEY")
    if api_key:
        return api_key

    # 6
    config = Config()
    api_cfg = config.get(api_name, {})
    if f"{api_name.lower()}_api_key" in api_cfg:
        api_key = api_cfg[f"{api_name.lower()}_api_key"]
        if api_key:
            return api_key

    # 7
    if "api_keys" in create_args and api_name in create_args["api_keys"]:
        api_key = create_args["api_keys"][api_name]
        if api_key:
            return api_key

    if strict:
        provider_upper = api_name.upper()
        api_key_env_var = f"{provider_upper}_API_KEY"
        api_key_arg = f"{api_name.lower()}_api_key"
        error_message = (
            f"API key for {api_name} not found. Please provide it using one of the following methods:\n"
            f"1. Set the {api_key_env_var} environment variable\n"
            f"2. Provide it as '{api_key_arg}' parameter or 'api_key' parameter when creating an agent using the CREATE AGENT syntax\n"
            f"   Example: CREATE AGENT my_agent USING model='gpt-4', provider='{api_name}', {api_key_arg}='your-api-key';\n"
            f"   Or: CREATE AGENT my_agent USING model='gpt-4', provider='{api_name}', api_key='your-api-key';\n"
        )
        raise Exception(error_message)
    return None



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


connection_args = OrderedDict(
    host={
        "type": ARG_TYPE.STR,
        "description": "The host name or IP address of the MonetDB server/database.",
    },
    database={
        "type": ARG_TYPE.STR,
        "description": """
            The database name to use when connecting with the MonetDB server.
        """,
    },
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate with the MonetDB server.",
    },
    password={
        "type": ARG_TYPE.PWD,
        "description": "The password to authenticate the user with the MonetDB server.",
        "secret": True,
    },
    port={
        "type": ARG_TYPE.INT,
        "description": "Specify port to connect MonetDB through TCP/IP",
    },
    schema_name={
        "type": ARG_TYPE.STR,
        "description": "Specify the schema name for Listing Table ",
    },
)

connection_args_example = OrderedDict(
    host="127.0.0.1",
    port=50000,
    password="monetdb",
    user="monetdb",
    schemaName="sys",
    database="demo",
)



import os
import pandas as pd
import openai
from openai import OpenAI, NotFoundError, AuthenticationError
from typing import Dict, Optional
from mindsdb.integrations.handlers.openai_handler import Handler as OpenAIHandler
from mindsdb.integrations.utilities.handler_utils import get_api_key
from mindsdb.integrations.handlers.groq_handler.settings import groq_handler_config
from mindsdb.utilities import log

logger = log.getLogger(__name__)


class GroqHandler(OpenAIHandler):
    """
    This handler handles connection to the Groq.
    """

    name = 'groq'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.api_base = groq_handler_config.BASE_URL
        self.default_model = groq_handler_config.DEFAULT_MODEL
        self.default_mode = groq_handler_config.DEFAULT_MODE
        self.supported_modes = groq_handler_config.SUPPORTED_MODES

    @staticmethod
    def _check_client_connection(client: OpenAI):
        """
        Check the Groq engine client connection by listing models.

        Args:
            client (OpenAI): OpenAI client configured with the Groq API credentials.

        Raises:
            Exception: If the client connection (API key) is invalid.

        Returns:
            None
        """
        try:
            client.models.list()
        except NotFoundError:
            pass
        except AuthenticationError as e:
            if e.body['code'] == 401:
                raise Exception('Invalid api key')
            raise Exception(f'Something went wrong: {e}')

    def create_engine(self, connection_args):
        """
        Validate the Groq API credentials on engine creation.

        Args:
            connection_args (dict): Connection arguments.

        Raises:
            Exception: If the handler is not configured with valid API credentials.

        Returns:
            None
        """
        connection_args = {k.lower(): v for k, v in connection_args.items()}
        api_key = connection_args.get('groq_api_key')
        if api_key is not None:
            org = connection_args.get('api_organization')
            api_base = connection_args.get('api_base') or os.environ.get('groq_BASE', groq_handler_config.BASE_URL)
            client = self._get_client(api_key=api_key, base_url=api_base, org=org)
            GroqHandler._check_client_connection(client)

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        """
        Validate the Groq API credentials on model creation.

        Args:
            target (str): Target column, not required for LLMs.
            args (dict): Handler arguments.
            kwargs (dict): Handler keyword arguments.

        Raises:
            Exception: If the handler is not configured with valid API credentials.

        Returns:
            None
        """
        if 'using' not in args:
            raise Exception(
                "Groq engine requires a USING clause! Refer to its documentation for more details."
            )
        else:
            args = args['using']

        engine_storage = kwargs['handler_storage']
        connection_args = engine_storage.get_connection_args()
        api_key = get_api_key('groq', args, engine_storage=engine_storage)
        api_base = connection_args.get('api_base') or args.get('api_base') or os.environ.get('GROQ_BASE', groq_handler_config.BASE_URL)
        org = args.get('api_organization')
        client = OpenAIHandler._get_client(api_key=api_key, base_url=api_base, org=org)
        GroqHandler._check_client_connection(client)

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:
        """
        Call the Groq engine to predict the next token.

        Args:
            df (pd.DataFrame): Input data.
            args (dict): Handler arguments.

        Returns:
            pd.DataFrame: Predicted data
        """
        api_key = get_api_key('groq', args, self.engine_storage)
        supported_models = self._get_supported_models(api_key, self.api_base)
        self.chat_completion_models = [model.id for model in supported_models]
        return super().predict(df, args)

    @staticmethod
    def _get_supported_models(api_key, base_url, org=None):
        """
        Get the list of supported models for the Groq engine.

        Args:
            api_key (str): API key.
            base_url (str): Base URL.
            org (str): Organization name.

        Returns:
            List: List of supported models.
        """
        client = openai.OpenAI(api_key=api_key, base_url=base_url, organization=org)
        return client.models.list()

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None):
        raise NotImplementedError("Fine-tuning is not supported for Groq AI engine.")
